###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

import sys
sys.path.append('/vistrails/src/trunk/vistrails')
import os
os.environ['EXECUTABLEPATH'] = ""
import copy
import datetime
import re
import urllib

from vistrails.db.domain import DBModule, DBConnection, DBPort, DBFunction, \
    DBParameter, DBLocation, DBPortSpec, DBTag, DBAnnotation, DBVistrail, \
    DBRegistry
import vistrails.db.services.io
from vistrails.db.services.io import SaveBundle
from vtk_imposter import vtk_module, vtk_function

class VTK2VT(object):
    vtk_data_url = 'http://www.vtk.org/files/release/5.0/vtkdata-5.0.4.zip'
    package_name = 'edu.utah.sci.vistrails.vtk'
    basic_name = 'edu.utah.sci.vistrails.basic'
    pythoncalc_name = 'edu.utah.sci.vistrails.pythoncalc'
    http_name = 'edu.utah.sci.vistrails.http'
    variant_sigstring ='(edu.utah.sci.vistrails.basic:Tuple)'
    unused_functions = set(['Update', 
                            'Read', 
                            'Write', 
                            'Start', 
                            'Initialize',
                            'Render', 
                            'PlaceWidget', 
                            'AddObserver',
                            'SetSize',
                            ])
    translate_ports = {'SetInputConnection': {None: 'SetInputConnection0'},
                       'GetOutputPort': {None: 'GetOutputPort0'},
                       'SetRenderWindow': {None: 'SetVTKCell'},
                       'SetInteractorStyle': {None: 'InteractorStyle'},
                       'ResetCamera': {None: 'ResetCamera'},
                       'AddInput': {None: 'AddInput',
                                    'vtkXYPlotActor': 'AddInput_2'},
                       'SetInput': {None: 'SetInput',
                                    'vtkPolyDataNormals': 'SetInput_1',
                                    'vtkGlyph3D': 'SetInput_1',
                                    'vtkDelaunay2D': 'SetInput_1',
                                    'vtkDelaunay3D': 'SetInput_1',
                                    'vtkWarpVector': 'SetInput_1',
                                    'vtkContourFilter': 'SetInput_1',
                                    'vtkTubeFilter': 'SetInput_1',
                                    'vtkThresholdPoints': 'SetInput_1',
                                    'vtkProbeFilter': 'SetInput_1',
                                    'vtkTriangleFilter': 'SetInput_1',
                                    'vtkBandedPolyDataContourFilter': \
                                        'SetInput_1',
                                    'vtkWarpScalar': 'SetInput_1'},
                       'SetSourceConnection': {None: 'SetSourceConnection',
                                               'vtkGlyph3D': \
                                                   'SetSourceConnection_2'},
                       'AddFunction': {None: 'AddFunction',
                                       ('vtkImplicitSum', 'vtkPlane'): \
                                           'AddFunction_2',
                                       ('vtkImplicitSum', 'Tuple'): \
                                           'AddFunction_1'},
                       'SetColor': {None: 'SetColor',
                                    ('vtkVolumeProperty', 
                                     'vtkColorTransferFunction'): \
                                        'SetColor_4',
                                    ('vtkVolumeProperty', 
                                     'vtkPiecewiseFunction'): \
                                        'SetColor_2'},
                       'SetScalarOpacity': {None: 'SetScalarOpacity',
                                            'vtkVolumeProperty': \
                                                'SetScalarOpacity_2'},
                       'SetSource': {None: 'SetSource',
                                     'vtkGlyph3D': 'SetSource_1'},
                       }

    inspector_functions = {'GetScalarRange': 'vtkDataSetInspector',
                           'GetCenter': 'vtkDataSetInspector',
                           'GetNumberOfCells': 'vtkDataSetInspector',
                           'GetNumberOfPoints': 'vtkDataSetInspector',
                           'GetLength': 'vtkDataSetInspector',
                           'GetBounds': 'vtkDataSetInspector',
                           'GetPointData': 'vtkDataSetInspector',
                           'GetCellData': 'vtkDataSetInspector',
                           'GetVerts': 'vtkPolyDataInspector',
                           'GetPoints': 'vtkPolyDataInspector',
                           'GetLines': 'vtkPolyDataInspector',
                           'GetStrips': 'vtkPolyDataInspector',
                           'GetPoints': 'vtkPolyDataInspector',
                           'GetNumberOfVerts': 'vtkPolyDataInspector',
                           'GetNumberOfPoints': 'vtkPolyDataInspector',
                           'GetNumberOfPolys': 'vtkPolyDataInspector',
                           'GetNumberOfStrips': 'vtkPolyDataInspector',
                           'GetScalars': 'vtkDataSetAttributesInspector',
                           'GetVectors': 'vtkDataSetAttributesInspector',
                           'GetNormals': 'vtkDataSetAttributesInspector',
                           'GetTCoords': 'vtkDataSetAttributesInspector',
                           'GetTensors': 'vtkDataSetAttributesInspector',
                           'GetGlobalIds': 'vtkDataSetAttributesInspector',
                           'GetPedigreeIds': 'vtkDataSetAttributesInspector',
                           'GetMaxNorm': 'vtkDataArrayInspector',
                           'GetRange': 'vtkDataArrayInspector',
                           }
                           
    getters = {'GetProperty': {None: ('SetProperty', 'vtkProperty', None),
                               'vtkXYPlotActor': ('SetProperty',
                                                  'vtkProperty2D', None)}, 
               'GetActiveCamera': {None: ('SetActiveCamera', 
                                          'vtkCamera', None)},
               'GetTitleTextProperty':{None: ('SetTitleTextProperty', 
                                              'vtkTextProperty',
                                              [('BoldOn',),
                                               ('ItalicOn',),
                                               ('ShadowOn',),
                                               ('SetFontFamilyToArial',)])},
               'GetTextProperty': {None: ('SetTextProperty', 
                                          'vtkTextProperty', None)},
               'GetPointData': {None: ('SetPointData', 'vtkPointData', None)},
               'GetCellData': {None: ('SetCellData', 'vtkCellData', None)},
               'GetPointIds': {None: ('SetPointIds', 'vtkIdList', None),
                               'vtkVoxel': ('SetPointIds', 'vtkIdList',
                                            [('SetNumberOfIds', 8)]),
                               'vtkHexahedron': ('SetPointIds', 'vtkIdList',
                                                 [('SetNumberOfIds', 8)]),
                               'vtkTetra': ('SetPointIds', 'vtkIdList',
                                            [('SetNumberOfIds', 4)]),
                               'vtkWedge': ('SetPointIds', 'vtkIdList',
                                            [('SetNumberOfIds', 6)]),
                               'vtkPyramid': ('SetPointIds', 'vtkIdList',
                                              [('SetNumberOfIds', 5)]),
                               'vtkPixel': ('SetPointIds', 'vtkIdList',
                                            [('SetNumberOfIds', 4)]), 
                               'vtkQuad': ('SetPointIds', 'vtkIdList',
                                           [('SetNumberOfIds', 4)]), 
                               'vtkTriangle': ('SetPointIds', 'vtkIdList',
                                               [('SetNumberOfIds', 3)]), 
                               'vtkLine': ('SetPointIds', 'vtkIdList',
                                           [('SetNumberOfIds', 2)]), 
                               'vtkVertex': ('SetPointIds', 'vtkIdList',
                                             [('SetNumberOfIds', 1)]), 
                               },
               'GetRenderWindow': {None: ('SetVTKCell', 'VTKCell', None)},
               }

    set_file_name_pattern = re.compile('Set.*FileName$')

    def __init__(self):
        self.created_modules = {}
        self.modules = {}
        self.vt_modules = {}
        self.vtk_modules = {}
        self.connections = {}
        self.vt_f_connections = {}
        self.vt_b_connections = {}
        self.max_module_id = 0
        self.max_location_id = 0
        self.max_function_id = 0
        self.max_parameter_id = 0
        self.max_connection_id = 0
        self.max_port_id = 0
        self.max_port_spec_id = 0
        self.http_module = None

        r_file = '/vistrails/registry.xml'
        self.registry = vistrails.db.services.io.open_registry_from_xml(r_file)
        self.registry_desc_idx = {}
        for package in self.registry.db_packages:
            for desc in package.db_module_descriptors:
                self.registry_desc_idx[desc.db_id] = desc

    def create_module(self, module):
        # print 'creating module', module._name, id(module)
        module_name = module._name
        package_name = self.package_name
        if module_name == 'vtkRenderWindow':
            module_name = 'VTKCell'
        elif module_name == 'vtkRenderWindowInteractor':
            return
        elif module_name == 'Tuple' or module_name == 'Unzip':
            package_name = self.basic_name
        elif module_name == 'PythonCalc':
            package_name = self.pythoncalc_name
        elif module_name == 'DownloadFile':
            package_name = self.http_name
        db_module = DBModule(id=self.max_module_id,
                             name=module_name,
                             package=package_name,
                             location=DBLocation(id=self.max_location_id, 
                                                 x=0.0, y=0.0),
                             )
        self.max_module_id += 1
        self.max_location_id += 1
        self.modules[id(module)] = db_module
        self.vt_modules[db_module.db_id] = db_module
        self.vtk_modules[id(module)] = module
        # self.vt_f_connections[db_module.db_id] = []
        # self.vt_b_connections[db_module.db_id] = []

    def save_modules(self):
        self.saved_functions = {}
        for k, module in self.vtk_modules.iteritems():
            self.saved_functions[k] = module.functions
            module.functions = []
            
    def restore_modules(self):
        for k, module in self.vtk_modules.iteritems():
            module.functions = self.saved_functions[k]
    
    def do_layout(self):
        queue = []
        # sinks = []
        vt_b_connections = self.vt_f_connections
        vt_f_connections = self.vt_b_connections
        b_connections = copy.deepcopy(vt_b_connections)
        f_connections = copy.deepcopy(vt_f_connections)
        for m_id, module in self.vt_modules.iteritems():
            if m_id not in b_connections:
                queue.append((module, 0))
            # if module.db_id not in f_connections:
            #     sinks.append([set([module])])

        # print >>sys.stderr, 'queue:', queue
        # topological sort
        t_sort = []
        for module, level in queue:
            t_sort.append((module, level))
            if module.db_id in f_connections:
                for (next_id, _) in f_connections[module.db_id]:
                    next_module = self.vt_modules[next_id]
                    if next_module.db_id in b_connections:
                        if len(b_connections[next_module.db_id]) == 0:
                            continue
                        remaining_list = b_connections[next_module.db_id][:]
                        for t in remaining_list:
                            if t[0] == module.db_id:
                                b_connections[next_module.db_id].remove(t)
                        if len(b_connections[next_module.db_id]) == 0:
                            queue.append((next_module, level+1))
        
        # print >>sys.stderr, [(m.db_name, ell) for (m, ell) in t_sort]
        levels = {}
        module_levels = {}
        for (m, ell) in t_sort:
            if ell not in levels:
                levels[ell] = []
            levels[ell].append(m)
            module_levels[m.db_id] = ell

        for level, m_list in sorted(levels.iteritems()):
            for i, m in enumerate(m_list):
                if i >= len(m_list) - 1:
                    continue

                m_next = m_list[i+1]
                m_idxs = []
                m_next_idxs = []
                if m.db_id not in vt_b_connections:
                    continue
                if m_next.db_id not in vt_b_connections:
                    continue
                for (n_id, _) in vt_b_connections[m.db_id]:
                    n = self.vt_modules[n_id]
                    n_idx = levels[module_levels[n_id]].index(n)
                    m_idxs.append(n_idx)
                for (n_id, _) in vt_b_connections[m_next.db_id]:
                    n = self.vt_modules[n_id]
                    n_idx = levels[module_levels[n_id]].index(n)
                    m_next_idxs.append(n_idx)
                num_overlaps_f = 0
                num_overlaps_b = 0
                for m_idx in m_idxs:
                    for m_next_idx in m_next_idxs:
                        if m_idx == m_next_idx:
                            continue
                        elif m_idx < m_next_idx:
                            num_overlaps_b += 1
                        else:
                            num_overlaps_f += 1
                if num_overlaps_f > num_overlaps_b:
                    # swap
                    swap = m_list[i]
                    m_list[i] = m_list[i+1]
                    m_list[i+1] = swap
                                
        dx = 250.0
        dy = 100.0
        current_y = 0.0
        for level, m_list in levels.iteritems():
            current_x = - (len(m_list) * dx) / 2.0
            for m in m_list:
                m.db_location.db_x = current_x
                m.db_location.db_y = current_y
                current_x += dx
            current_y += dy
                   
    def get_vtk_cell(self, interactor):
        found_vtk_cell = False
        if interactor._name == 'vtkRenderWindowInteractor':
            found_vtk_cell = False
            for function in interactor.functions:
                if function._name == 'SetRenderWindow' and \
                        len(function._args) == 1 and \
                        type(function._args[0]) == vtk_module:
                    # use VTKCell to translate calls
                    function_list = interactor.functions
                    module = function._args[0]
                    found_vtk_cell = True
                    break
        if found_vtk_cell:
            return (module, function_list)
        else:
            print >>sys.stderr, "ERROR: cannot find VTKCell for Interactor"
        return (None, None)
    
    def create_functions(self, module, function_list=None, f_pos=0):
        # print >>sys.stderr, 'creating functions:', module._name
#         import traceback
#         traceback.print_stack(file=sys.stderr)
        if module._name == 'vtkRenderWindowInteractor':
            (module, new_function_list) = self.get_vtk_cell(module)
            if not module:
                # punt on it in this case
                return
            else:
                function_list = []
                for function in new_function_list[:]:
                    if type(function.parent) == vtk_module:
                        function.parent = module
                    if function._name != 'SetRenderWindow':
                        function_list.append(function)
        if function_list is None:
            function_list = module.functions
        for function in function_list:
            # print >>sys.stderr, '  creating:', function._name
            db_function = self.create_function(function, f_pos)
            if db_function:
                self.modules[id(module)].db_add_function(db_function)
                f_pos += 1
        return f_pos

    def find_or_create_module(self, function, allow_sub=True):
#         print >>sys.stderr, 'find_or_create:', function._name, \
#             function.parent._name, id(function)
        parent_name = function.parent._name

#                    'GetPositionCoordinate': {None: ('SetPosition',
#                                                     'vtkCoordinate')},
#                    'GetPosition2Coordinate': {None: ('SetPosition2',
#                                                      'vtkCoordinate')}
#                    }
        
        # if we have a getter, than check this stuff
        if function._name.startswith('Get') and len(function._args) == 0 and \
                len(function.sub_functions) == 1 and \
                function.sub_functions[0]._name in self.inspector_functions:
            # print >>sys.stderr, ">> found inspector function"
            # print >>sys.stderr, "  ", function._name, function.parent._name
            inspector_name = \
                self.inspector_functions[function.sub_functions[0]._name]
            key = (id(function.parent), function._name)
            # print >>sys.stderr, '  ', key

            out_function = function.sub_functions[0]
            if key in self.created_modules:
                inspector_module = self.created_modules[key]
            else:
                inspector_module = vtk_module(inspector_name)
                self.create_module(inspector_module)
                self.created_modules[key] = inspector_module

                # need to attach first piece of call to inspector input
                in_function = vtk_function('SetInput')
                in_function.parent = inspector_module
                function.sub_functions = []
                in_function._args = [function]
                inspector_module.functions.append(in_function)
                self.create_functions(inspector_module, [in_function])
            # need to attach rest to inspector output
            found = False
            out_function.parent = inspector_module
            for old_function in inspector_module.functions:
                if old_function._name == out_function._name and \
                        (len(old_function.sub_functions) == 0 or
                         len(out_function.sub_functions) == 0 or
                         old_function.sub_functions[0]._name == \
                             old_function.sub_functions[0]._name):
                    found = True
                    break
            if not found:
                inspector_module.functions.append(out_function)
                self.create_functions(inspector_module, [out_function])
            return (inspector_module, True)
        elif function._name in self.getters and \
                (function._name not in self.inspector_functions or
                 self.inspector_functions[function._name] != \
                     function.parent._name):
#             and len(function.sub_functions) == 1 \
#                 and function.sub_functions[0]._name not in \
#                 self.inspector_functions:
            # print >>sys.stderr, '>>', function._name, 'found in getters'
            if not allow_sub:
                print >>sys.stderr, "ERROR: sub function has getter"
            key = (id(function.parent), function._name)
            if key in self.created_modules:
                p_module = self.created_modules[key]
#                 print >>sys.stderr, "found in created_modules:", \
#                     p_module._name
                return (p_module, False)
            if parent_name in self.getters[function._name]:
                (setter, module_name, function_list) = \
                    self.getters[function._name][parent_name]
            else:
                (setter, module_name, function_list) = \
                    self.getters[function._name][None]
#             print >>sys.stderr, 'self.getters:', self.getters[function._name]
            found_setter = False
            for p_function in function.parent.functions:
                if p_function._name == setter:
                    found_setter = True
                    for arg in p_function._args:
                        if type(arg) == vtk_module:
                            p_module = arg
                        else:
                            if arg == function:
                                found_setter = False
                            else:
                                p_module = arg.parent
                    if found_setter and type(p_module) != vtk_module:
                        print >>sys.stderr, "ERROR: type of p_module incorrect"
                        found_setter = False
                    if found_setter:
                        # print >>sys.stderr, "found", module_name, p_module._name
                        self.create_functions(p_module, function.sub_functions)
                        return (p_module, False)
            if not found_setter:
                # print >>sys.stderr, "doing create for", module_name
                p_module = \
                    self.create_module_for_get(function, setter, module_name,
                                               function_list)
                # print >>sys.stderr, id(p_module)
            return (p_module, False)
        return (None, False)

    def create_module_for_get(self, function, setter, module_name, 
                              function_list=None):
        # create new vtkProperty module
        # add connection to parent module on SetProperty call
        # for sub_function in function.sub_functions:
            # add each sub_function to the vtkProperty module
        src_module = None
        # if len(function.sub_functions) > 0:

        db_module = DBModule(id=self.max_module_id,
                     name=module_name,
                     package=self.package_name,
                     location=DBLocation(id=self.max_location_id, 
                                         x=0.0, y=0.0),
                     )
        self.max_module_id += 1
        self.max_location_id += 1
        src_module = vtk_module(module_name)                
        self.modules[id(src_module)] = db_module
        self.vt_modules[db_module.db_id] = db_module
        self.vtk_modules[id(src_module)] = src_module
        if function_list:
            src_functions = []
            for f in function_list:
                # doesn't work for functions with args yet
                src_function = vtk_function(f[0])
                if len(f) > 1:
                    src_function._args = list(f[1:])
                src_function.parent = src_module
                src_functions.append(src_function)
                # print >>sys.stderr, '**', src_function._name
            # src_module.functions = src_functions
            self.create_functions(src_module, src_functions)
            # print >>sys.stderr, "** done with create"

        # self.vt_f_connections[db_module.db_id] = []
        # self.vt_b_connections[db_module.db_id] = []
        self.created_modules[(id(function.parent), function._name)] = \
            src_module
        dst_function = vtk_function(setter)
        dst_function.parent = function.parent
        self.create_connection(src_module, dst_function)

        return src_module

    def create_tuple(self, function):
        tuple_module = vtk_module('Tuple')
        tuple_module.blah()
        self.create_module(tuple_module)
        db_tuple_module = self.modules[id(tuple_module)]
        tuple_f_pos = 0
        final_sigs = []
        for i, arg in enumerate(function._args):
            port_spec = DBPortSpec(id=self.max_port_spec_id,
                                   name='arg' + str(i),
                                   type='input',
                                   sigstring=self.variant_sigstring)
            db_tuple_module.db_add_portSpec(port_spec)
            self.max_port_spec_id += 1
            if type(arg) == vtk_module or \
                    type(arg) == vtk_function:
                # connection
                tuple_function = vtk_function('arg' + str(i))
                tuple_function.parent = tuple_module
                conn = self.create_connection(arg, tuple_function)
                port_spec.db_sigstring = \
                    conn.db_ports_type_index['source'].db_signature
                conn.db_ports_type_index['destination'].db_signature = \
                    port_spec.db_sigstring
                final_sigs.append(port_spec.db_sigstring[1:-1])
            else:
                db_parameters = self.get_parameters([arg])
                db_function = DBFunction(id=self.max_function_id,
                                         name='arg' + str(i),
                                         pos=tuple_f_pos,
                                         parameters=db_parameters,
                                         )
                self.max_function_id += 1
                tuple_f_pos += 1
                db_tuple_module.db_add_function(db_function)
                port_spec.db_sigstring = \
                    "(" + ','.join(p.db_type for p in db_parameters) + ")"
                final_sigs.extend(p.db_type for p in db_parameters)
        final_sigstring = "(" + ",".join(final_sigs) + ")"
        port_spec = DBPortSpec(id=self.max_port_spec_id,
                               name='value',
                               type='output',
                               sigstring=final_sigstring)
        db_tuple_module.db_add_portSpec(port_spec)
        self.max_port_spec_id += 1
        tuple_output = vtk_function('value')
        tuple_output.parent = tuple_module
        self.create_connection(tuple_output, function)
        return tuple_module

    def get_parameters(self, params):
        def get_type(arg):
            python_types = {int: 'edu.utah.sci.vistrails.basic:Integer', 
                            str: 'edu.utah.sci.vistrails.basic:String', 
                            float: 'edu.utah.sci.vistrails.basic:Float',
                            bool: 'edu.utah.sci.vistrails.basic:Boolean'}
            if type(arg) not in python_types:
                return None
            return python_types[type(arg)]
        # end get_type

        def create_param(arg, position):
            p_type = get_type(arg)
            if p_type:
                p = DBParameter(id=self.max_parameter_id,
                            type=p_type,
                            val=arg,
                            pos=position,
                            )
                db_parameters.append(p)
                self.max_parameter_id += 1
                position += 1
            return position
        # end create_param

        db_parameters = []
        pos = 0
        for arg in params:
            if type(arg) == tuple:
                for tuple_arg in arg:
                    pos = create_param(tuple_arg, pos)
            else:
                pos = create_param(arg, pos)
        return db_parameters
    # end get_parameters
        

    def create_function(self, function, f_pos, allow_sub=True):

        # print >>sys.stderr, 'type(function.parent):', type(function.parent)
        db_function = None
        (p_module, _) = self.find_or_create_module(function, allow_sub)
        if p_module:
            # print >>sys.stderr, 'found p_module', p_module._name, 'for', function._name
            add_functions = []
            for sub_function in function.sub_functions:
                if not sub_function._name.startswith('Get'):
                    # print >>sys.stderr, 'adding function:', sub_function._name
                    add_functions.append(sub_function)
            self.create_functions(p_module, add_functions, 
                                  len(self.modules[id(p_module)].db_functions))
#             self.create_functions(p_module, function.sub_functions,
#                                   len(self.modules[id(p_module)].db_functions))
            db_function = None
        else:
            if function._name in self.unused_functions:
                return None
            parameters = []
            connections = []
            for arg in function._args:
                if type(arg) == vtk_module or \
                        type(arg) == vtk_function:                    
                    connections.append(arg)
                else:
                    parameters.append(arg)

            if len(function._args) == 0 and \
                    len(function.sub_functions) == 0 and \
                    not function._name.startswith('Get'):
                parameters.append(None)
            elif function._name.startswith('Get'):
                parameters = []
                connections = []
#             if function._name == 'SetScalarRange' and \
#                     len(function._args) == 1:
#                 # use a vtkDataSetInspector to get scalar range
#                 arg = function._args[0]
#                 if type(arg) == vtk_function and \
#                         type(arg.parent) == vtk_function and \
#                         arg._name == 'GetScalarRange' and \
#                         arg.parent._name == 'GetOutput':
#                     inspector_module = vtk_module('vtkDataSetInspector')
#                     inspector_module.SetInputConnection(
#                         arg.parent.parent.GetOutputPort())
#                     self.create_module(inspector_module)
#                     self.create_functions(inspector_module)

#                     function._args = (inspector_module.GetScalarRange(),)
#                     connections = [function._args[0]]
#             elif function._name == 'GenerateValues' and \
#                     len(function._args) == 2:
#                 # use a vtkDataSetInspector to get scalar range
#                 arg = function._args[0]
#                 if type(arg) != int:
#                     print >>sys.stderr, \
#                         "ERROR: GenerateValues first arg type is incorrect"
#                     return None
#                 arg = function._args[1]
#                 if type(arg) == vtk_function and \
#                         type(arg.parent) == vtk_function and \
#                         arg._name == 'GetScalarRange' and \
#                         arg.parent._name == 'GetOutput':
#                     inspector_module = vtk_module('vtkDataSetInspector')
#                     inspector_module.SetInputConnection(
#                         arg.parent.parent.GetOutputPort())
#                     self.create_module(inspector_module)
#                     self.create_functions(inspector_module)

#                     function._args = (function._args[0], 
#                                       inspector_module.GetScalarRange())
#                     connections = [function._args[1]]
#                     parameters = [function._args[0]]   
            if self.set_file_name_pattern.match(function._name) and \
                    type(function.parent) == vtk_module and \
                    len(function._args) == 1 and \
                    type(function._args[0]) == str:
                # replace SetFileName with a DownloadFile -> Unzip -> SetFile
                if not self.http_module:
                    self.http_module = vtk_module('DownloadFile')
                    self.http_module.url(self.vtk_data_url)
                    self.create_module(self.http_module)
                    self.create_functions(self.http_module)
                unzip_module = vtk_module('Unzip')
                self.create_module(unzip_module)
                unzip_module.filename_in_archive(function._args[0][1:])
                unzip_module.archive_file(self.http_module.file())
                self.create_functions(unzip_module)
                set_file_function = vtk_function(function._name[:-4])
                set_file_function.parent = function.parent
                set_file_function._args = [unzip_module.file()]
                self.create_functions(function.parent, [set_file_function])
                return None
            if function._name == 'SetInteractor' and \
                    len(function._args) == 1 and \
                    type(function._args[0]) == vtk_module and \
                    type(function.parent) == vtk_module:
                observer_module = function.parent
                for a_function in observer_module.functions:
                    # right way to do this is to make SharedData a Tuple
                    if a_function._name == 'AddObserver' and \
                            len(a_function._args) == 2: 
                        (vtk_cell, _) = self.get_vtk_cell(function._args[0])
                        handler_module = vtk_module('vtkInteractionHandler')
                        handler_module.Observer(function.parent)
                        self.create_module(handler_module)
                        vtk_cell.InteractionHandler(handler_module)
                        for b_index, b_function in \
                                enumerate(vtk_cell.functions):
                            if b_function._name == 'InteractionHandler':
                                break
                        # probably need to make this position right...
                        self.create_functions(vtk_cell, [b_function],
                                              len(self.modules[id(vtk_cell)].db_functions))
                        del vtk_cell.functions[b_index]
                     
                        # print >>sys.stderr, a_function._args
                        (event_type, python_function) = a_function._args
                        # print >>sys.stderr, 'AddObserver', event_type, \
                        #     'executing'
                        self.save_modules()
                        python_function(observer_module, None)
                        f = event_type[0].lower() + event_type[1:]
                        f = f.replace('Event', 'Handler')
                        handler_code = 'def %s(observer, shareddata):\n' % f
                        # probably need to order the function calls here
                        # ideally, this would be an abstraction of some sort
                        shared_data = set()
                        for s_module in self.vtk_modules.itervalues():
                            if len(s_module.functions) > 0:
                                if not s_module._name == 'vtkRenderWindow':
                                    shared_data.add(s_module)
                                    # print >>sys.stderr, s_module._name
                                    for s_function in s_module.functions:
                                        for s_arg in s_function._args:
                                            while type(s_arg) == vtk_function:
                                                s_arg = s_arg.parent
                                            if type(s_arg) == vtk_module:
                                                shared_data.add(s_arg)
                                                #print >>sys.stderr, s_arg._name

                        s_codelines = []
                        s_arg_map = {}
                        s_arg_idx = 1

                        if len(shared_data) < 1 or (len(shared_data) == 1 and \
                                                        observer_module in \
                                                        shared_data):
                            handler_code += '    pass\n'
                        else:
                            if len(shared_data) > 2:
                                handler_code += '    ('
                                comma = ','
                            else:
                                handler_code += '    '
                                comma = ''
                            for s_module in shared_data:
                                if s_module != observer_module:
                                    handler_code += 'arg' + str(s_arg_idx) + \
                                        comma
                                    s_arg_map[id(s_module)] = s_arg_idx
                                    loop_arg_idx = s_arg_idx
                                    s_arg_idx += 1
                                else:
                                    loop_arg_idx = 0
                                for s_function in s_module.functions:
                                    s_line = (s_function.id, loop_arg_idx, 
                                              s_function)
                                    # print >>sys.stderr, "s_line:", s_line
                                    s_codelines.append(s_line)
                            if len(shared_data) > 2:
                                handler_code += ') = shareddata\n'
                            else:
                                handler_code += ' = shareddata\n'
                            # need to assign input module names, then write code
                            s_codelines.sort()

                            # need to check to see if an arg...
                            for (_, m_idx, s_function) in s_codelines:
                                s_module_name = 'observer' if m_idx == 0 else \
                                    'arg%d.vtkInstance' % m_idx
                                s_function_str = s_function._name
                                while len(s_function.sub_functions) > 0:
                                    if len(s_function.sub_functions) > 1:
                                        print >> sys.stderr, \
                                            "ERROR: too many subfunctions"
                                    # FIXME need to do args here, too
                                    s_function_str += '().' + \
                                        s_function.sub_functions[0]._name
                                    s_function = s_function.sub_functions[0]
                                handler_code += '    %s.%s(' % \
                                    (s_module_name, s_function_str)

                                s_arg_str = ''
                                for arg in s_function._args:
                                    if type(arg) == vtk_module:
                                        s_module = arg
                                    elif type(arg) == vtk_function:
                                        s_module = arg.parent
                                    else:
                                        s_arg_str += arg + ', '
                                        continue
                                    m_idx = s_arg_map[id(s_module)]
                                    s_module_name = 'observer' if m_idx == 0 \
                                        else 'arg%d.vtkInstance' % m_idx
                                    if type(arg) == vtk_module:
                                        s_arg_str += s_module_name
                                    else:
                                        s_arg_str += s_module_name + \
                                            arg._name + '()'
                                handler_code += s_arg_str + ')\n'
                                    
                                    
                        self.restore_modules()                        
                        # print >>sys.stderr, handler_code
#                         s_data_function = vtk_function('SharedData')
#                         s_data_function.parent = handler_module
#                         s_data_function._args = list(shared_data)
                        shared_data.discard(observer_module)
                        shared_data.discard(handler_module)
                        handler_module.SharedData(*shared_data)
                        handler_module.Handler(urllib.quote(handler_code))
                        # print >> sys.stderr, "shared_data", shared_data
                        # functions.append(s_data_function)
                        self.create_functions(handler_module)

                return None
            elif len(function.sub_functions) == 1:
                sub_function = function.sub_functions[0]
                if function._name == 'GetPositionCoordinate' and \
                        sub_function._name == 'SetValue':
                    # print >>sys.stderr, 'sub_function:', sub_function._args[:2]
                    new_function = vtk_function('SetPosition')
                    new_function._args = sub_function._args[:2]
                    new_function._kwargs = sub_function._kwargs
                    new_function.parent = function.parent
                    function = new_function
                    parameters.extend(new_function._args)
                elif function._name == 'GetPosition2Coordinate' and \
                        sub_function._name == 'SetValue':
                    # print >>sys.stderr, 'sub_function:', sub_function._args[:2]
                    new_function = vtk_function('SetPosition2')
                    new_function._args = sub_function._args[:2]
                    new_function._kwargs = sub_function._kwargs
                    new_function.parent = function.parent
                    function = new_function
                    parameters.extend(new_function._args)
                    
#                 # we need to use a Tuple module
#                 tuple_module = vtk_module('Tuple')
#                 self.create_module(tuple_module)
#                 new_arg_list = []
#                 for arg in function._args:
#                     if type(arg) == vtk_module or \
#                             type(arg) == vtk_function:
#                         self.create_connection(arg, tuple_module)
#                     else:
#                         new_arg_list.append(arg)
#                 tuple_function = vtk_function(
#                 function._args = 

            if len(connections) > 0:
                if len(connections) > 1 or len(parameters) > 0:
                    # need tuple
                    tuple_module = self.create_tuple(function)
                    db_function = None
                else:
                    if allow_sub:
                        for arg in connections:
                            self.create_connection(arg, function)
                    else:
                        print >>sys.stderr, \
                            "ERROR: sub function has connections"
            elif len(parameters) > 0:
                db_parameters = self.get_parameters(parameters)
                db_function = DBFunction(id=self.max_function_id,
                                         name=function._name,
                                         pos=f_pos,
                                         parameters=db_parameters,
                                         )
                self.max_function_id += 1
        return db_function
        
    def create_connection(self, src, dst):
        convert_type_map = {'input': 'destination', 'output': 'source'}
        def create_port(vtk_module, type, name, other_vtk_module):
            # print >>sys.stderr, 'creating port:', vtk_module._name, name
            if id(vtk_module) not in self.modules:
                print >>sys.stderr, 'ERROR: cannot create port', \
                    vtk_module._name, type, name
                import traceback
                traceback.print_stack()
                return (None, None)

            db_module = self.modules[id(vtk_module)]
            db_package = \
                self.registry.db_packages_identifier_index[db_module.db_package]
            if db_module.db_namespace is None:
                db_module.db_namespace = ''
            d_key = (db_module.db_name, db_module.db_namespace)
            db_module_desc = db_package.db_module_descriptors_name_index[d_key]

            if name in self.translate_ports:
                if db_module.db_name in self.translate_ports[name]:
                    name = self.translate_ports[name][db_module.db_name]
                elif (db_module.db_name, 
                      other_vtk_module._name) in self.translate_ports[name]:
                    name = self.translate_ports[name][(db_module.db_name,
                                                       other_vtk_module._name)]
                else:
                    name = self.translate_ports[name][None]

            ps_key = (name, type)
            if ps_key in db_module.db_portSpecs_name_index:
                db_module_desc = db_module
            while ps_key not in db_module_desc.db_portSpecs_name_index:
                base_id = db_module_desc.db_base_descriptor_id
                if base_id < 0:
                    print >>sys.stderr, 'ERROR: cannot create port', \
                        vtk_module._name, type, name
                    import traceback
                    traceback.print_stack()
                    return (None, None)
                db_module_desc = self.registry_desc_idx[base_id]

            db_port_spec = db_module_desc.db_portSpecs_name_index[ps_key]
            db_signature = db_port_spec.db_sigstring

            new_port = DBPort(id=self.max_port_id,
                              type=convert_type_map[type],
                              moduleId=db_module.db_id,
                              moduleName=db_module.db_name,
                              name=name,
                              signature=db_signature,
                              )
            self.max_port_id += 1
            return (db_module, new_port)
        # end create_port
        
        def find_port_info(entity):
            if type(entity) == vtk_module:
                if entity._name == 'vtkRenderWindowInteractor':
                    (entity, _) = self.get_vtk_cell(entity)
                found_module = entity
                port_name = 'self'
            elif type(entity) == vtk_function and \
                    type(entity.parent) == vtk_module:
                (found_module, inspector) = self.find_or_create_module(entity)
                if found_module:
                    if inspector:
                        port_name = entity._name
                    else:
                        port_name = 'self'
                else:
                    found_module = entity.parent
                    port_name = entity._name
            elif type(entity) == vtk_function and  \
                    type(entity.parent) == vtk_function:
                (found_module, _) = self.find_or_create_module(entity.parent)
                port_name = entity._name
                if not found_module:
                    print >>sys.stderr, \
                        "ERROR: cannot convert function", entity._name
                    
#             elif type(entity) == vtk_function:
#                 found_module = self.find_or_create_module(entity)
#                 if found_module:
#                     port_name = 'self'
#                 else:
#                     if type(entity.parent) == vtk_function:
#                         found_module = self.find_or_create_module(entity.parent)
#                         port_name = entity._name
#                         if not found_module:
#                             print >>sys.stderr, \
#                                 "ERROR: cannot convert function", entity._name
#                     else:
#                         found_module = entity.parent
#                         port_name = entity._name
            else:
                print >>sys.stderr, 'ERROR: cannot deal with type "' + \
                    str(type(entity))
                found_module = None
                port_name = None
            return (found_module, port_name)
        # end find_port_info

        src_port = None
        src_module = None
        dst_port = None
        dst_module = None

        # print >> sys.stderr, "creating connection:", src._name, dst._name

        (src_vtk_module, src_port_name) = find_port_info(src)
        (dst_vtk_module, dst_port_name) = find_port_info(dst)
        if src_vtk_module and dst_vtk_module:
            (src_module, src_port) = create_port(src_vtk_module, 'output',
                                                 src_port_name, dst_vtk_module)
            (dst_module, dst_port) = create_port(dst_vtk_module, 'input',
                                                 dst_port_name, src_vtk_module)

#         if type(src) == vtk_module:
#             src_vtk_module = src
#             (src_module, src_port) = create_port(src, 'source', 'self', 
#                                                  dst.parent)
#         elif type(src) == vtk_function:
#             # print >>sys.stderr, 'src:', src._name, src.parent._name
#             src_vtk_module = self.find_or_create_module(src)
#             if src_vtk_module:
#                 print >>sys.stderr, '** found early "%s"' % src._name, src_vtk_module._name
#                 print >>sys.stderr, 'src_vtk_module:', src_vtk_module._name
#                 src_name = 'self'
#             if not src_vtk_module:
#                 if type(src.parent) == vtk_function:
#                     src_vtk_module = self.find_or_create_module(src.parent)
#                     src_name = src._name
#                     print >>sys.stderr, 'found src_vtk_module', src_vtk_module._name, 'for', src.parent._name, src_name

#                     # print >>sys.stderr, "src_vtk_module:", id(src_vtk_module)
#                     if not src_vtk_module:
#                         print >>sys.stderr, \
#                             "ERROR: cannot convert function", src._name
#                 else:
#                     src_vtk_module = src.parent
#                     src_name = src._name
#             if src_vtk_module:
#                 (src_module, src_port) = create_port(src_vtk_module, 'source', 
#                                                      src_name, dst.parent)
#         else:
#             print >>sys.stderr, 'ERROR: cannot deal with type "' + \
#                 str(type(src)) + '"'

#         if type(dst) == vtk_function:
#             (dst_module, dst_port) = create_port(dst.parent, 'destination',
#                                                  dst._name, src_vtk_module)
#         else:
#             print >>sys.stderr, 'ERROR: cannot deal with type "' + \
#                 str(type(dst)) + '"'

        if not src_module or not dst_module:
            print >>sys.stderr, 'ERROR: cannot create connection', \
                src._name, dst._name
            return None
#         print >>sys.stderr, 'creating connection:', \
#             src_module.db_name, dst_module.db_name
        connection = DBConnection(id=self.max_connection_id,
                                  ports=[src_port, dst_port],
                                  )
        self.max_connection_id += 1
        key = (id(src_module), id(dst_module))
        if key not in self.connections:
            self.connections[key] = []
        self.connections[key].append(connection)
        if src_module.db_id not in self.vt_f_connections:
            self.vt_f_connections[src_module.db_id] = []
        self.vt_f_connections[src_module.db_id].append((dst_module.db_id, 
                                                        connection))
        if dst_module.db_id not in self.vt_b_connections:
            self.vt_b_connections[dst_module.db_id] = []
        self.vt_b_connections[dst_module.db_id].append((src_module.db_id,
                                                        connection))
        return connection

    def print_vt(self, filename=None):
        import vistrails.db.services.io
        import vistrails.db.services.action

        action_list = []
        for module in self.modules.itervalues():
            # print 'module:', module._name, id(module)
            action_list.append(('add', module))
        for connection_list in self.connections.itervalues():
            for connection in connection_list:
                # print 'connection:', connection._name, id(connection)
                action_list.append(('add', connection))
        self.do_layout()
        action = vistrails.db.services.action.create_action(action_list)
        action.db_id = 1
        action.db_prevId = 0
        action.db_user = 'dakoop'
        action.db_date = datetime.datetime.now()

        tag = DBTag(name='Auto-Translated', id=1L)
            
        op_id = 0
        for op in action.db_operations:
            op.db_id = op_id
            op_id += 1
        vistrail = DBVistrail(actions=[action], tags=[tag], version='0.9.0')
        if filename: 
            # f = file(filename, 'w')
            script_name = os.path.splitext(filename)[0] + '.py'

            # total hack to make this a capital E
            script_name = 'E' + script_name[3:]
            annotation = DBAnnotation(id=1,
                                      key='notes',
                                      value="This workflow was automatically generated from a modified version of the vtk python example script '%s' from the vtk 5.0.4 distribution.  In most cases, running this workflow will generate a visualization that is identical to the result from the example script." % script_name)
            action.db_add_annotation(annotation)

            vistrails.db.services.io.save_bundle_to_zip_xml(SaveBundle(DBVistrail.vtType, vistrail), filename)
        else:
            f = None
            print >>f, vistrails.db.services.io.serialize(vistrail)
            



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

""" The file describes the parameter tree view that allows changing the alias
of parameters

QAliasParameterView
"""
from __future__ import division

from PyQt4 import QtCore, QtGui
from PyQt4.QtCore import pyqtSignal, pyqtSlot
from vistrails.core.inspector import PipelineInspector
from vistrails.core.vistrail.module_param import ModuleParam
from vistrails.gui.common_widgets import QSearchTreeWindow, QSearchTreeWidget
from vistrails.gui.paramexplore.pe_pipeline import QAnnotatedPipelineView
from vistrails.gui.vistrails_palette import QVistrailsPaletteInterface
from vistrails.gui.theme import CurrentTheme
from vistrails.core.utils import InstanceObject

################################################################################
class QAliasParameterView(QtGui.QWidget, QVistrailsPaletteInterface):
    
    def __init__(self, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.set_title("Mashup Pipeline")
        layout = QtGui.QVBoxLayout()
        self.parameter_panel = QAliasParameterPanel()
        font = QtGui.QFont("Arial", 11, QtGui.QFont.Normal)
        font.setItalic(True)
        label = QtGui.QLabel("Double-click a parameter to change alias")
        label.setFont(font)
        param_group = QtGui.QGroupBox(self.parameter_panel.windowTitle())
        g_layout = QtGui.QVBoxLayout()
        g_layout.setMargin(0)
        g_layout.setSpacing(2)
        g_layout.addWidget(label)
        g_layout.addWidget(self.parameter_panel)
        param_group.setLayout(g_layout)
        layout.addWidget(param_group)
        
        self.pipeline_view = QAnnotatedPipelineView()
        self.pipeline_view.setReadOnlyMode(True)
        p_view_group = QtGui.QGroupBox(self.pipeline_view.windowTitle())
        g_layout = QtGui.QVBoxLayout()
        g_layout.setMargin(0)
        g_layout.setSpacing(0)
        g_layout.addWidget(self.pipeline_view)
        p_view_group.setLayout(g_layout)
        layout.addWidget(p_view_group)
        self.setLayout(layout)
        
        self.parameter_panel.treeWidget.aliasChanged.connect(self.aliasChanged)
        
    def updateMshpController(self, mshpController):
        from vistrails.gui.vistrails_window import _app
        self.mshpController = mshpController
        self.parameter_panel.set_pipeline(
                        self.mshpController.vtPipeline)
        self.pipeline_view.set_controller(self.mshpController.vtController)
        self.mshpController.vtController.current_pipeline_view = \
                                                            self.pipeline_view
        self.pipeline_view.scene().current_pipeline = self.mshpController.vtPipeline
        self.mshpController.vtController.current_pipeline = self.mshpController.vtPipeline
        
        #print "**** should update mashup pipeline view "
        
        #self.pipeline_view.scene().setupScene(self.mshpController.vtPipeline)
        self.pipeline_view.scene().clear()
        self.pipeline_view.version_changed()
        self.pipeline_view.zoomToFit()
        self.pipeline_view.updateAnnotatedIds(
                        self.mshpController.vtPipeline)
        #_app.notify('mashup_pipeline_view_set')
        
    def updateMshpVersion(self, version):
        #print "will update alias param view"
        self.parameter_panel.set_pipeline(
                        self.mshpController.vtPipeline)
        self.pipeline_view.version_changed()
        
    def zoomToFit(self):
        if self.pipeline_view:
            self.pipeline_view.zoomToFit()
            
    def aliasChanged(self, param):
        from vistrails.gui.vistrails_window import _app
        _app.notify('alias_changed', param)
        
################################################################################

class QAliasParameterPanel(QSearchTreeWindow):
    """
    QAliasParameterPanel is a special widget for displaying methods and 
    parameters inside a pipeline that allow changing aliases
    
    """
    def createTreeWidget(self):
        """ createTreeWidget() -> QModuleTreeWidget
        Return the search tree widget for this window
        
        """
        self.setWindowTitle("Parameter Aliases")
        return QAliasParameterTreeWidget(self)

    def set_pipeline(self, pipeline):
        self.pipeline = pipeline
        self.treeWidget.updateFromPipeline(pipeline)
        
################################################################################
class QAliasParameterTreeWidget(QSearchTreeWidget):
    """
    QAliasParameterTreeWidget is a subclass of QSearchTreeWidget to display all
    Vistrails Module
    
    """
    #signals
    aliasChanged = pyqtSignal(InstanceObject)
    
    def __init__(self, parent=None):
        """ QAliasParameterTreeWidget(parent: QWidget) -> QParameterTreeWidget
        Set up size policy and header

        """
        QSearchTreeWidget.__init__(self, parent)
        self.header().hide()
        self.setRootIsDecorated(False)
        self.delegate = QAliasParameterTreeWidgetItemDelegate(self, self)
        self.setItemDelegate(self.delegate)
        self.aliasNames = []
        self.itemDoubleClicked.connect(self.changeAlias)
    def updateFromPipeline(self, pipeline):
        """ updateFromPipeline(pipeline: Pipeline) -> None
        Read the list of aliases and parameters from the pipeline
        
        """
        self.clear()
        if not pipeline:
            return

        # Update the aliases
#        if len(pipeline.aliases)>0:
#            aliasRoot = QParameterTreeWidgetItem(None, self, ['Aliases'])
#            aliasRoot.setFlags(QtCore.Qt.ItemIsEnabled,
#                               )
#            for (alias, info) in pipeline.aliases.iteritems():
#                ptype, pId, parentType, parentId, _ = info
#                parameter = pipeline.db_get_object(ptype, pId)
#                v = parameter.strValue
#                aType = parameter.type
#                aIdentifier = parameter.identifier
#                aNamespace = parameter.namespace
#                label = ['%s = %s' % (alias, v)]
#                pInfo = InstanceObject(type=aType,
#                                      identifier=aIdentifier,
#                                      namespace=aNamespace,
#                                      value=v,
#                                      id=pId,
#                                      dbtype=ptype,
#                                      parent_dbtype=parentType,
#                                      parent_id=parentId,
#                                      is_alias=True)
#                aliasItem = QParameterTreeWidgetItem((alias, [pInfo]),
#                                                     aliasRoot, label)
#            aliasRoot.setExpanded(True)
            
        # Now go through all modules and functions
        self.aliasNames = pipeline.aliases.keys()
        inspector = PipelineInspector()
        inspector.inspect_ambiguous_modules(pipeline)
        sortedModules = sorted(pipeline.modules.iteritems(),
                               key=lambda item: item[1].name)
        for mId, module in sortedModules:
            if len(module.functions)>0:
                mLabel = [module.name]
                moduleItem = None
                for fId in xrange(len(module.functions)):
                    function = module.functions[fId]
                    if len(function.params)==0: continue
                    if moduleItem==None:
                        if inspector.annotated_modules.has_key(mId):
                            annotatedId = inspector.annotated_modules[mId]
                            moduleItem = QAliasParameterTreeWidgetItem(annotatedId,
                                                                       self, mLabel)
                        else:
                            moduleItem = QAliasParameterTreeWidgetItem(None,
                                                                  self, mLabel)
                    #v = ', '.join([p.strValue for p in function.params])
                    label = ['%s'%function.name]
                    
                    pList = [InstanceObject(type=function.params[pId].type,
                                           identifier=function.params[pId].identifier,
                                           namespace=function.params[pId].namespace,
                                           value=function.params[pId].strValue,
                                           id=function.params[pId].real_id,
                                           dbtype=ModuleParam.vtType,
                                           parent_dbtype=function.vtType,
                                           parent_id=function.real_id,
                                           alias=function.params[pId].alias,
                                           mId=mId)
                             for pId in xrange(len(function.params))]
                    mName = module.name
                    if moduleItem.parameter!=None:
                        mName += '(%d)' % moduleItem.parameter
                    fName = '%s :: %s' % (mName, function.name)
                    mItem = QAliasParameterTreeWidgetItem((fName, pList),
                                                     moduleItem,
                                                     label)
                if moduleItem:
                    moduleItem.setExpanded(True)
                    
    @pyqtSlot(QtGui.QTreeWidgetItem, int)    
    def changeAlias(self, item, column ):
        """ itemClicked(item:  , column: int) -> None        
        If mouse click on the item, show up a dialog to change/add
        the alias name
        
        """
        if isinstance(item.parameter, InstanceObject):
            (text, ok) = QtGui.QInputDialog.getText(self,
                                                    'Set Parameter Alias',
                                                    'Enter the parameter alias',
                                                    QtGui.QLineEdit.Normal,
                                                    item.parameter.alias)
            while ok and str(text) in self.aliasNames:
                msg =" This alias is already being used.\
 Please enter a different parameter alias "
                (text, ok) = QtGui.QInputDialog.getText(self,
                                                        'Set Parameter Alias',
                                                        msg,
                                                        QtGui.QLineEdit.Normal,
                                                        text)
            if ok and item.parameter.alias != str(text):
                item.parameter.alias = str(text)
                item.updateAlias()
                self.aliasChanged.emit(item.parameter)
            
            
        
class QAliasParameterTreeWidgetItemDelegate(QtGui.QItemDelegate):
    """    
    QAliasParameterTreeWidgetItemDelegate will override the original
    QTreeWidget paint function to draw buttons for top-level item
    similar to QtDesigner. This mimics
    Qt/tools/designer/src/lib/shared/sheet_delegate, which is only a
    private class from QtDesigned.
    
    """
    def __init__(self, view, parent):
        """ QParameterTreeWidgetItemDelegate(view: QTreeView,
                                          parent: QWidget)
                                          -> QParameterTreeWidgetItemDelegate
        Create the item delegate given the tree view
        
        """
        QtGui.QItemDelegate.__init__(self, parent)
        self.treeView = view

    def paint(self, painter, option, index):
        """ painter(painter: QPainter, option QStyleOptionViewItem,
                    index: QModelIndex) -> None
        Repaint the top-level item to have a button-look style
        
        """
        model = index.model()
        if (model.parent(index).isValid()==False or 
            model.parent(index).parent().isValid()==False):
            style = self.treeView.style()
            r = option.rect
            textrect = QtCore.QRect(r.left() + 10,
                                    r.top(),
                                    r.width() - 10,
                                    r.height())
            font = painter.font()
            font.setBold(True)
            painter.setFont(font)
            text = option.fontMetrics.elidedText(
                model.data(index, QtCore.Qt.DisplayRole),
                QtCore.Qt.ElideMiddle, 
                textrect.width()-10)
            style.drawItemText(painter,
                               textrect,
                               QtCore.Qt.AlignLeft,
                               option.palette,
                               self.treeView.isEnabled(),
                               text)
            painter.setPen(QtGui.QPen(QtCore.Qt.black))
            fm = QtGui.QFontMetrics(font)
            size = fm.size(QtCore.Qt.TextSingleLine, text)
            #painter.drawLine(textrect.left()-5,
            #                 textrect.bottom()-1,
            #                 textrect.left()+size.width()+5,
            #                 textrect.bottom()-1)

            annotatedId = model.data(index, QtCore.Qt.UserRole+1)            
            if annotatedId:
                idRect = QtCore.QRect(
                    QtCore.QPoint(textrect.left()+size.width()+5,
                                  textrect.top()),
                    textrect.bottomRight())
                QAnnotatedPipelineView.drawId(painter, idRect,
                                              annotatedId,
                                              QtCore.Qt.AlignLeft |
                                              QtCore.Qt.AlignVCenter)
        else:
            QtGui.QItemDelegate.paint(self, painter, option, index)

    def sizeHint(self, option, index):
        """ sizeHint(option: QStyleOptionViewItem, index: QModelIndex) -> None
        Take into account the size of the top-level button
        
        """
        return (QtGui.QItemDelegate.sizeHint(self, option, index) +
                QtCore.QSize(2, 2))
            

class QAliasParameterTreeWidgetItem(QtGui.QTreeWidgetItem):
    """
    QParameterTreeWidgetItem represents module on QParameterTreeWidget
    
    """
    def __init__(self, info, parent, labelList):
        """ QAliasParameterTreeWidgetItem(info: (str, []),
                                     parent: QTreeWidgetItem
                                     labelList: string)
                                     -> QParameterTreeWidget
                                     
        Create a new tree widget item with a specific parent and
        labels. info describing a set of parameters as follow:
        (name, [InstanceObject]):
           name  = Name of the parameter set (alias or function)
        If this item is a top-level item, info can either be None or
        an integer specifying the annotated id of this module

        """
        self.parameter = info
        QtGui.QTreeWidgetItem.__init__(self, parent, labelList)
        if isinstance(self.parameter, int):
            self.setData(0, QtCore.Qt.UserRole+1, self.parameter)
        elif isinstance(self.parameter, tuple):
            for param in self.parameter[1]:
                label = ['']
                item = QAliasParameterTreeWidgetItem(param, self, label)
                item.setFlags(QtCore.Qt.ItemIsEnabled|
                              QtCore.Qt.ItemIsSelectable)
                item.updateAlias()
            self.setExpanded(True)
    
    def updateAlias(self):    
        if isinstance(self.parameter, InstanceObject):
            if self.parameter.alias != '':
                self.setText(0,'%s(%s):%s'%(self.parameter.alias, 
                                          self.parameter.type,
                                          self.parameter.value))
                self.setIcon(0,CurrentTheme.MASHUP_ALIAS_ICON)
            else:
                self.setText(0,'%s:%s'%(self.parameter.type,
                                      self.parameter.value))
                self.setIcon(0,QtGui.QIcon())



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division, with_statement

import datetime
import functools
import locale
import os
import platform
import sys
import time
import urllib2
import warnings

from vistrails.core import debug
from vistrails.core.utils import unimplemented, VistrailsDeprecation, Chdir


###############################################################################

from common import *

def with_c_locale(func):
    @functools.wraps(func)
    def newfunc(*args, **kwargs):
        previous_locale = locale.setlocale(locale.LC_TIME)
        locale.setlocale(locale.LC_TIME, 'C')
        try:
            return func(*args, **kwargs)
        finally:
            locale.setlocale(locale.LC_TIME, previous_locale)
    return newfunc

@with_c_locale
def strptime(*args, **kwargs):
    """Version of datetime.strptime that always uses the C locale.

    This is because date strings are used internally in the database, and
    should not be localized.
    """
    return datetime.datetime.strptime(*args, **kwargs)

@with_c_locale
def time_strptime(*args, **kwargs):
    """Version of time.strptime that always uses the C locale.

    This is because date strings are used internally in the database, and
    should not be localized.
    """
    return time.strptime(*args, **kwargs)

@with_c_locale
def strftime(dt, *args, **kwargs):
    """Version of datetime.strftime that always uses the C locale.

    This is because date strings are used internally in the database, and
    should not be localized.
    """
    if hasattr(dt, 'strftime'):
        return dt.strftime(*args, **kwargs)
    else:
        return time.strftime(dt, *args, **kwargs)

##############################################################################

systemType = platform.system()

if systemType in ['Windows', 'Microsoft']:
    from vistrails.core.system.windows import *

elif systemType in ['Linux']:
    from vistrails.core.system.linux import *

elif systemType in ['Darwin']:
    from vistrails.core.system.osx import *
else:
    debug.critical("VisTrails could not detect your operating system.")
    sys.exit(1)

###############################################################################

# Makes sure root directory is sensible.
if __name__ == '__main__':
    _thisDir = sys.argv[0]
else:
    _thisDir = sys.modules[__name__].__file__
_thisDir = os.path.split(_thisDir)[0]
__rootDir = os.path.realpath(os.path.join(_thisDir,
                                          '..',
                                          '..'))

__dataDir = os.path.realpath(os.path.join(__rootDir,
                                          'data'))
__fileDir = os.path.realpath(os.path.join(__rootDir,
                                          '..','examples'))

if systemType in ['Darwin'] and not os.path.exists(__fileDir):
    # Assume we are running from py2app
    __fileDir = os.path.realpath(os.path.join(__rootDir,
                                              '/'.join(['..']*6),'examples'))

__examplesDir = __fileDir

__defaultFileType = '.vt'

_defaultPkgPrefix = 'org.vistrails.vistrails'

def get_vistrails_default_pkg_prefix():
    """Gets the namespace under which identifiers of builtin packages live.

    You should *not* use this, it is only useful intended to expand short names
    of builtin packages in parse_descriptor_string.
    """
    warnings.warn("get_vistrails_default_pkg_prefix() is deprecated",
                  category=VistrailsDeprecation)
    return _defaultPkgPrefix

def get_vistrails_basic_pkg_id():
    return "%s.basic" % _defaultPkgPrefix

def get_vistrails_directory(config_key, conf=None):
    if conf is None:
        from vistrails.core.configuration import get_vistrails_configuration
        conf = get_vistrails_configuration()
    if conf.has_deep_value(config_key):
        d = conf.get_deep_value(config_key)
        if os.path.isabs(d):
            return d
        else:
            return os.path.join(current_dot_vistrails(conf), d)
    return None

def set_vistrails_data_directory(d):
    """ set_vistrails_data_directory(d:str) -> None
    Sets vistrails data directory taking into account environment variables

    """
    global __dataDir
    new_d = os.path.expanduser(d)
    new_d = os.path.expandvars(new_d)
    while new_d != d:
        d = new_d
        new_d = os.path.expandvars(d)
    __dataDir = os.path.realpath(d)

def set_vistrails_file_directory(d):
    """ set_vistrails_file_directory(d: str) -> None
    Sets vistrails file directory taking into accoun environment variables

    """
    global __fileDir
    new_d = os.path.expanduser(d)
    new_d = os.path.expandvars(new_d)
    while new_d != d:
        d = new_d
        new_d = os.path.expandvars(d)
    __fileDir = os.path.realpath(d)

def set_vistrails_root_directory(d):
    """ set_vistrails_root_directory(d:str) -> None
    Sets vistrails root directory taking into account environment variables

    """

    global __rootDir
    new_d = os.path.expanduser(d)
    new_d = os.path.expandvars(new_d)
    while new_d != d:
        d = new_d
        new_d = os.path.expandvars(d)
    __rootDir = os.path.realpath(d)

def set_vistrails_default_file_type(t):
    """ set_vistrails_default_file_type(t:str) -> None
    Which file type to use when the user doesn't provide a file extension

    """
    global __defaultFileType
    if t in ['.vt', '.xml']:
        __defaultFileType = t
    else:
        __defaultFileType = '.vt'

def vistrails_root_directory():
    """ vistrails_root_directory() -> str
    Returns vistrails root directory

    """
    return __rootDir

def vistrails_file_directory():
    """ vistrails_file_directory() -> str 
    Returns current vistrails file directory

    """
    return __fileDir

def vistrails_examples_directory():
    """ vistrails_file_directory() -> str 
    Returns vistrails examples directory

    """
    return __examplesDir

def vistrails_data_directory():
    """ vistrails_data_directory() -> str
    Returns vistrails data directory

    """
    return __dataDir

def vistrails_default_file_type():
    """ vistrails_default_file_type() -> str
    Returns vistrails file type

    """
    return __defaultFileType

def packages_directory():
    """ packages_directory() -> str
    Returns vistrails packages directory

    """
    return os.path.join(vistrails_root_directory(),'packages')

def blank_vistrail_file():
    unimplemented()

def resource_directory():
    """ resource_directory() -> str
    Returns vistrails gui resource directory

    """
    return os.path.join(vistrails_root_directory(),
                        'gui', 'resources')

def default_options_file():
    """ default_options_file() -> str
    Returns vistrails default options file

    """
    return os.path.join(home_directory(), ".vistrailsrc")

def default_dot_vistrails():
    """ default_dot_vistrails() -> str
    Returns the default VisTrails per-user directory.

    """
    return os.path.join(home_directory(), '.vistrails')

def current_dot_vistrails(conf=None):
    """ current_dot_vistrails() -> str
    Returns the VisTrails per-user directory.

    """
    if conf is None:
        from vistrails.core.configuration import get_vistrails_configuration
        conf = get_vistrails_configuration()
    return conf.dotVistrails

def default_connections_file():
    """ default_connections_file() -> str
    Returns default Vistrails per-user connections file

    """
    return os.path.join(current_dot_vistrails(), 'connections.xml')

VERSION = '2.2'
def vistrails_version():
    """vistrails_version() -> string - Returns the current VisTrails version."""
    # 0.1 was the Vis2005 version
    # 0.2 was the SIGMOD demo version
    # 0.3 was the plugin/vtk version
    # 0.4 is cleaned up version with new GUI
    # 1.0 is version with new schema
    return VERSION

def get_latest_vistrails_version():
    """get_latest_vistrails_version() -> string - Returns latest vistrails
    release version as queried from vistrails.org"""

    version = ''
    version_url = \
            "http://www.vistrails.org/download/download.php?id=release_version.txt"
    try:
        request = urllib2.Request(version_url)
        get_latest_version = urllib2.urlopen(request)
        version = get_latest_version.read().strip()
    except urllib2.HTTPError, err:
        debug.warning("Unable to check for updates: %s" % str(err))
        return version

    return version

def new_vistrails_release_exists():
    """ new_vistrail_release_exists() -> (bool, str)
    Returns (True, new_version_str) if newer version exists

    """

    local_version = [int(x) for x in vistrails_version().split('.')]

    remote_str = get_latest_vistrails_version()
    if remote_str:
        remote_version = [int(x) for x in remote_str.split('.')]
    else:
        remote_version = [0]

    if cmp(local_version, remote_version) is -1:
        return (True, remote_str)

    return (False, None)

def vistrails_revision():
    """vistrails_revision() -> str
    When run on a working copy, shows the current svn revision else
    shows the latest release revision

    """
    git_dir = os.path.join(vistrails_root_directory(), '..')
    with Chdir(git_dir):
        release = vistrails_version()
        import vistrails.core.requirements
        if vistrails.core.requirements.executable_file_exists('git'):
            lines = []
            result = execute_cmdline(
                ['git', 'describe', '--always'],
                lines)
            if len(lines) == 1:
                if result == 0:
                    release = lines[0].strip(" \n")
    return release


_registry = None
def get_module_registry():
    global _registry
    if _registry is None:
        from vistrails.core.modules.module_registry import get_module_registry
        _registry = get_module_registry()
    return _registry

def short_about_string():
    return """VisTrails version %s (%s) -- contact@vistrails.org""" % \
            (vistrails_version(), vistrails_revision())

def about_string():
    """about_string() -> string - Returns the about string for VisTrails."""
    return """VisTrails version %s (%s) -- contact@vistrails.org

Copyright (C) 2014-2015 New York University. Copyright (C) 2011-2014 NYU-Poly.
Copyright (C) 2006-2011 University of Utah.
All rights reserved.
http://www.vistrails.org

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the New York University nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.""" % (vistrails_version(),
                                                         vistrails_revision())

###############################################################################

import unittest

if __name__ == '__main__':
    unittest.main()

class TestSystem(unittest.TestCase):

    def test_vistrails_revision(self):
        r = vistrails_root_directory()
        with Chdir(r):
            v1 = vistrails_revision()
            try:
                with Chdir(os.path.join(r, '..')):
                    self.assertEquals(v1, vistrails_revision())
            except AssertionError:
                raise
            except Exception:
                pass
            try:
                with Chdir(os.path.join(r, '..', '..')):
                    self.assertEquals(v1, vistrails_revision())
            except AssertionError:
                raise
            except Exception:
                pass



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'module_appearance.ui'
#
# Created: Mon Oct 11 20:15:11 2010
#      by: PyQt4 UI code generator 4.7.2
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(819, 602)
        self.centralwidget = QtGui.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayoutWidget = QtGui.QWidget(self.centralwidget)
        self.verticalLayoutWidget.setGeometry(QtCore.QRect(0, 0, 811, 551))
        self.verticalLayoutWidget.setObjectName("verticalLayoutWidget")
        self.verticalLayout = QtGui.QVBoxLayout(self.verticalLayoutWidget)
        self.verticalLayout.setObjectName("verticalLayout")
        self.graphicsView = QtGui.QGraphicsView(self.verticalLayoutWidget)
        self.graphicsView.setObjectName("graphicsView")
        self.verticalLayout.addWidget(self.graphicsView)
        self.horizontalLayout_2 = QtGui.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.textEdit = QtGui.QTextEdit(self.verticalLayoutWidget)
        self.textEdit.setObjectName("textEdit")
        self.horizontalLayout_2.addWidget(self.textEdit)
        self.textEdit_help = QtGui.QTextEdit(self.verticalLayoutWidget)
        self.textEdit_help.setObjectName("textEdit_help")
        self.horizontalLayout_2.addWidget(self.textEdit_help)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.horizontalLayout = QtGui.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.pushButton_symmetric = QtGui.QPushButton(self.verticalLayoutWidget)
        self.pushButton_symmetric.setObjectName("pushButton_symmetric")
        self.horizontalLayout.addWidget(self.pushButton_symmetric)
        self.pushButton_mirrored = QtGui.QPushButton(self.verticalLayoutWidget)
        self.pushButton_mirrored.setObjectName("pushButton_mirrored")
        self.horizontalLayout.addWidget(self.pushButton_mirrored)
        self.pushButton_Clear = QtGui.QPushButton(self.verticalLayoutWidget)
        self.pushButton_Clear.setObjectName("pushButton_Clear")
        self.horizontalLayout.addWidget(self.pushButton_Clear)
        spacerItem = QtGui.QSpacerItem(40, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.pushButton_Quit = QtGui.QPushButton(self.verticalLayoutWidget)
        self.pushButton_Quit.setObjectName("pushButton_Quit")
        self.horizontalLayout.addWidget(self.pushButton_Quit)
        self.verticalLayout.addLayout(self.horizontalLayout)
        MainWindow.setCentralWidget(self.centralwidget)
        self.statusbar = QtGui.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QtGui.QApplication.translate("MainWindow", "MainWindow", None, QtGui.QApplication.UnicodeUTF8))
        self.textEdit_help.setHtml(QtGui.QApplication.translate("MainWindow", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Sans\'; font-size:10pt; font-weight:400; font-style:normal;\">\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">To edit the shape, double-click the fringe lines and drag the vertices. To delete the vertices, select them and press &quot;Delete&quot; on your keyboard. </p>\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">The resulting code snippet can be used to generate a module fringe in VisTrails like the one being displayed.</p>\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">The buttons &quot;Make rotationally symmetric&quot; and &quot;Make mirrored&quot; use the left side of the module to produce the right side. They <span style=\" font-weight:600;\">will erase</span> the decorations on the right side.</p></body></html>", None, QtGui.QApplication.UnicodeUTF8))
        self.pushButton_symmetric.setText(QtGui.QApplication.translate("MainWindow", "Make rotationally symmetric", None, QtGui.QApplication.UnicodeUTF8))
        self.pushButton_mirrored.setText(QtGui.QApplication.translate("MainWindow", "Make mirrored", None, QtGui.QApplication.UnicodeUTF8))
        self.pushButton_Clear.setText(QtGui.QApplication.translate("MainWindow", "Clear", None, QtGui.QApplication.UnicodeUTF8))
        self.pushButton_Quit.setText(QtGui.QApplication.translate("MainWindow", "Quit", None, QtGui.QApplication.UnicodeUTF8))




###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
from __future__ import division

from vistrails.core.configuration import get_vistrails_persistent_configuration, \
    get_vistrails_configuration
from vistrails.gui.open_db_window import QOpenDBWindow, QConnectionDBSetupWindow
from vistrails.core.db.locator import DBLocator, FileLocator
from vistrails.db import VistrailsDBException
from vistrails.core import debug
import vistrails.db.services.io
from PyQt4 import QtGui, QtCore
import vistrails.core.system
import os

##############################################################################
# DB dialogs

def get_load_db_locator_from_gui(parent, obj_type):
    config, obj_id, obj_name = QOpenDBWindow.getOpenDBObject(obj_type)
    if config == {} or obj_id == -1:
        return None
    return DBLocator(config['host'],
                     config['port'],
                     config['db'],
                     config['user'],
                     config['passwd'],
                     obj_name,
                     obj_id=obj_id,
                     obj_type=obj_type,
                     connection_id=config.get('id', None))

def get_save_db_locator_from_gui(parent, obj_type, locator=None):
    config, name = QOpenDBWindow.getSaveDBObject(obj_type)
    if config == {} or name == '':
        return None
    return DBLocator(config['host'],
                     config['port'],
                     config['db'],
                     config['user'],
                     config['passwd'],
                     name,
                     obj_id=None,
                     obj_type=obj_type,
                     connection_id=config.get('id', None))

##############################################################################

def get_db_connection_from_gui(parent, id, name, host, port, user, passwd,
                               database):
    def show_dialog(parent, id, name, host, port, user,
                    passwd, databaseb, create):
        dialog = QConnectionDBSetupWindow(parent, id, name, host, port, user,
                                          passwd, database, create)
        config = None
        if dialog.exec_() == QtGui.QDialog.Accepted:
            config = {'host': str(dialog.hostEdt.text()),
                      'port': int(dialog.portEdt.value()),
                      'user': unicode(dialog.userEdt.text()),
                      'passwd': unicode(dialog.passwdEdt.text()),
                      'db': str(dialog.databaseEdt.text())
                      }
            try:
                vistrails.db.services.io.test_db_connection(config)
                config['succeeded'] = True
                config['name'] = str(dialog.nameEdt.text())
                config['id'] = dialog.id
            except VistrailsDBException, e:
                debug.critical('VisTrails DB Exception',  e)
                config['succeeded'] = False
        return config
    #check if the information is already there
    dbwindow = QOpenDBWindow.getInstance()
        
    config = dbwindow.connectionList.findConnectionInfo(host,port,database)

    if config:
        testconfig = dict(config)
        del testconfig['id']
        del testconfig['name']
        try:
            vistrails.db.services.io.test_db_connection(testconfig)
            config['succeeded'] = True
        except VistrailsDBException, e:
            config = show_dialog(parent, config['id'],
                                 config['name'], host, port, config['user'],
                                 passwd, database, create = False)
            
    elif config is None:
        config = show_dialog(parent, -1,"",
                             host, port, user, passwd,
                             database, create = True)
        if config['succeeded'] == True:
            #add to connection list
            dbwindow.connectionList.setConnectionInfo(**config)
    return config

##############################################################################
# File dialogs

suffix_map = {'vistrail': ['.vt', '.xml', '.vtl'],
              'workflow': ['.xml'],
              'log': ['.xml'],
              'registry': ['.xml'],
              'opm_graph': ['.xml'],
              'prov_document': ['.xml']
              }

def get_load_file_locator_from_gui(parent, obj_type):
    suffixes = "*" + " *".join(suffix_map[obj_type])
    fileName = QtGui.QFileDialog.getOpenFileName(
        parent,
        "Open %s..." % obj_type.capitalize(),
        vistrails.core.system.vistrails_file_directory(),
        "VisTrails files (%s)\nOther files (*)" % suffixes)
    if not fileName:
        return None
    filename = os.path.abspath(str(QtCore.QFile.encodeName(fileName)))
    dirName = os.path.dirname(filename)
    setattr(get_vistrails_persistent_configuration(), 'fileDir', dirName)
    setattr(get_vistrails_configuration(), 'fileDir', dirName)
    vistrails.core.system.set_vistrails_file_directory(dirName)
    return FileLocator(filename)

def get_save_file_locator_from_gui(parent, obj_type, locator=None):
    # Ignore current locator for now
    # In the future, use locator to guide GUI for better starting directory

    suffixes = "*" + " *".join(suffix_map[obj_type])
    fileName = QtGui.QFileDialog.getSaveFileName(
        parent,
        "Save Vistrail...",
        vistrails.core.system.vistrails_file_directory(),
        filter="VisTrails files (%s)" % suffixes, # filetypes.strip()
        options=QtGui.QFileDialog.DontConfirmOverwrite)
    if not fileName:
        return None
    f = str(QtCore.QFile.encodeName(fileName))

    # check for proper suffix
    found_suffix = False
    for suffix in suffix_map[obj_type]:
        if f.endswith(suffix):
            found_suffix = True
            break
    if not found_suffix:
        if obj_type == 'vistrail':
            f += get_vistrails_configuration().defaultFileType
        else:
            f += suffix_map[obj_type][0]

    if os.path.isfile(f):
        msg = QtGui.QMessageBox(QtGui.QMessageBox.Question,
                                "VisTrails",
                                "File exists. Overwrite?",
                                (QtGui.QMessageBox.Yes |
                                 QtGui.QMessageBox.No),
                                parent)
        if msg.exec_() == QtGui.QMessageBox.No:
            return None
    dirName = os.path.dirname(f)
    setattr(get_vistrails_persistent_configuration(), 'fileDir', dirName)
    setattr(get_vistrails_configuration(), 'fileDir', dirName)
    vistrails.core.system.set_vistrails_file_directory(dirName)
    return FileLocator(f)
   
def get_autosave_prompt(parent):
    """ get_autosave_prompt(parent: QWidget) -> bool
    
    """
    result = QtGui.QMessageBox.question(parent, 
                                        "AutoSave",
                                        "Autosave data has been found.\nDo you want to open autosave data?",
                                        QtGui.QMessageBox.Open,
                                        QtGui.QMessageBox.Ignore)
    return result == QtGui.QMessageBox.Open

def ask_to_overwrite_file(parent=None, obj_type='vistrail'):
    overwrite = True
    fname = None
    msg = QtGui.QMessageBox(QtGui.QMessageBox.Question,
                            "VisTrails",
                            "File exists and contains changes. Overwrite?",
                            (QtGui.QMessageBox.Yes | QtGui.QMessageBox.No),
                            parent)
    if msg.exec_() == QtGui.QMessageBox.No:
        overwrite = False
        locator = get_save_file_locator_from_gui(parent, obj_type)
        if locator:
            fname = locator._name
    return (overwrite, fname)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

class vtk_base(object):
    def __init__(self, n=None, *args, **kwargs):
        self._name = n
        self._args = args
        self._kwargs = kwargs
        
    def __call__(self, *args, **kwargs):
        # print 'calling %s __call__ with' % self.__class__.__name__, \
        #     args, kwargs
        return self.__class__.__dict__['_call'](self, self.n, *args, **kwargs)
        
    def __getattr__(self, name):
        # print 'calling %s __getattr__ with' % self.__class__.__name__, name
        if name in ('__str__','__repr__'): 
            return lambda: 'instance of %s at %s' % (str(self.__class__), 
                                                     id(self))
        self.n = name
        return self

    def _call(self, n, *args, **kwargs):
        # print 'calling _call with', n, args, kwargs
        pass

class vtk_function(vtk_base):
    id_count = 0

    def __init__(self, n=None, parent=None, *args, **kwargs):
        vtk_base.__init__(self, n, *args, **kwargs)
        self.parent = parent
        self.sub_functions = []
        self.id = vtk_function.id_count
        vtk_function.id_count += 1

    def _call(self, n, *args, **kwargs):
        vtk_base._call(self, n, *args, **kwargs)
        new_sub_function = vtk_function(n, self, *args, **kwargs)
        self.__dict__['sub_functions'].append(new_sub_function)
        return new_sub_function

class vtk_module(vtk_base):
    def __init__(self, n=None, *args, **kwargs):
        vtk_base.__init__(self, n, *args, **kwargs)
        self.functions = []

    def _call(self, n, *args, **kwargs):
        vtk_base._call(self, n, *args, **kwargs)
        new_function = vtk_function(n, self, *args, **kwargs)
        self.__dict__['functions'].append(new_function)
        return new_function
    
class vtk(vtk_base):

    def __init__(self, n=None, *args, **kwargs):
        vtk_base.__init__(self, n, *args, **kwargs)
        self.modules = []
        # print 'self.__dict__:', self.__dict__

    def _call(self, n, *args, **kwargs):
        vtk_base._call(self, n, *args, **kwargs)
        new_module = vtk_module(n, *args, **kwargs)
        self.__dict__['modules'].append(new_module)
        return new_module

    def to_vt(self, filename):
        from vtk_to_vt import VTK2VT
        my_writer = VTK2VT()
        for module in self.modules:
            my_writer.create_module(module)
        for module in self.modules:
            my_writer.create_functions(module)
        my_writer.print_vt(filename)



#!/usr/bin/env python
###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
"""This will write text of a release notes file based on commit messages and trac
tickets. Use the configuration section below to tune the start and end commits.
The text will be written to the standard output.
"""

import xmlrpclib
import git
import getpass
import os
import sys
import re
import tempfile
import subprocess
import shutil

#### configuration ####
commit_start = "269e4808eca3" # hash of version used on last release notes
commit_end = "HEAD" # current hash
branch = "v2.1" # git branch to be used
release_name = "2.1.4"
clonepath = None    # set this to the complete path of a vistrails clone to be
                    # used
                    # if None, the remote repository will be cloned to a
                    # temporary folder and removed at the end of the script
#clonepath = '/Users/tommy/git/vistrails'
cloneremote = 'git://www.vistrails.org/vistrails.git'
#### end configuration #####

## The script will ask for your Trac user and password
## so no need to change this now
username = None
password = None
need_cleanup = False

################################################################################

def userpass(realm, u, may_save):
    global username
    global password
    if not username:
        print "Username:",
        sys.stdout.flush()
        username = raw_input()
        password = getpass.getpass()
    return True, username, password, False

################################################################################

def rstrip(s):
    return '\n'.join(l.rstrip() for l in s.splitlines())

################################################################################

def clone_vistrails_git_repository(path_to):
    global cloneremote
    cmdlist = ['git', 'clone', cloneremote,
               path_to]
    print "Cloning vistrails from:"
    print "  %s to"%cloneremote
    print "  %s"%path_to
    print "Be patient. This may take a while."
    process = subprocess.Popen(cmdlist, shell=False,
                               stdin=subprocess.PIPE,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT)
    process.wait()
    if process.returncode == 0:
        print "repository is cloned."
    return process.returncode

################################################################################

def init_repo():
    global clonepath, need_cleanup, branch
    ok = False
    if clonepath is None:
        clonepath = tempfile.mkdtemp(prefix="vtrel")
        try:
            if clone_vistrails_git_repository(clonepath) == 0:
                ok = True
                init_branch(clonepath,branch)
            need_cleanup = True
        except Exception, e:
            print "ERROR: Could not clone vistrails repository!"
            print str(e)
            shutil.rmtree(clonepath)
            sys.exit(1)
    else:
        init_branch(clonepath,branch)
        pull_changes(clonepath)
        ok = True
    if ok:
        repo = git.Repo(clonepath)
        return repo
    else:
        print "ERROR: git clone failed."
        sys.exit(1)

################################################################################

def cleanup_repo():
    global clonepath, need_cleanup
    if need_cleanup:
        shutil.rmtree(clonepath)

################################################################################

def init_branch(path_to, branch):
    cmdlist = ['git', 'checkout', "%s"%branch]
    print "Checking out %s branch..."%branch
    current_dir = os.getcwd()
    os.chdir(path_to)
    process = subprocess.Popen(cmdlist, shell=False,
                               stdin=subprocess.PIPE,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT)
    process.wait()
    lines = process.stdout.readlines()
    for line in lines:
        print "   ", line
    if process.returncode == 0:
        print "Branch %s was checked out."%branch
    os.chdir(current_dir)
    return process.returncode

################################################################################

def pull_changes(path_to):
    cmdlist = ['git', 'pull']
    print "Pulling changes into the branch..."
    current_dir = os.getcwd()
    os.chdir(path_to)
    process = subprocess.Popen(cmdlist, shell=False,
                               stdin=subprocess.PIPE,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT)
    process.wait()
    lines = process.stdout.readlines()
    for line in lines:
        print "   ", line
    if process.returncode == 0:
        print "Changes were pulled."
    os.chdir(current_dir)
    return process.returncode

################################################################################

def checkout_branch(repo, branch):
    repobranch = getattr(repo.heads, branch)
    repobranch.checkout()

##############################################################################

def build_release_notes(repo, branch):
    global username
    global password
    global commit_start, commit_end

    def check_inside_skip(skip_list, message):
        found = False
        for s in skip_list:
            if s.find(message) != -1:
                found = True
                break
        return found

    re_ticket_old = re.compile(r'<ticket>(.*?)</ticket>', re.M | re.S)
    re_ticket = re.compile(r'^Ticket: (.*?)$', re.M | re.S)
    re_ticket2 = re.compile(r'^Fixes: (.*?)$', re.M | re.S)
    re_bugfix_old = re.compile(r'<bugfix>(.*?)</bugfix>', re.M | re.S)
    re_bugfix = re.compile(r'^Bugfix: (.*?)$', re.M | re.S)
    re_feature_old = re.compile(r'<feature>(.*?)</feature>', re.M | re.S)
    re_feature = re.compile(r'^Feature:(.*?)$', re.M | re.S)
    re_skip = re.compile(r'<skip>(.*?)</skip>', re.M | re.S)

    #build list and dictionary with commits
    logs = []
    log_map_time = {}
    checkout_branch(repo,branch)
    for c in repo.iter_commits("%s..%s"%(commit_start,commit_end)):
        logs.append(c)
        log_map_time[c.hexsha] = c.committed_date

    #populate dictionaries
    bugfixes = {}
    tickets = {}
    features = {}
    changes = {}
    ticket_info = {}

    for log in logs:
        ls = re_skip.findall(log.message)
        lf = re_feature.findall(log.message)
        lf.extend(re_feature_old.findall(log.message))
        lt = re_ticket.findall(log.message)
        lt.extend(re_ticket2.findall(log.message))
        lt.extend(re_ticket_old.findall(log.message))
        lb = re_bugfix.findall(log.message)
        lb.extend(re_bugfix_old.findall(log.message))
        for s in ls:
            changes[s.strip()] = log.hexsha
        for f in lf:
            features[f.strip()] = log.hexsha
        for t in lt:
            # handle tickets with # (should not be used)
            t = t.strip()
            if t.startswith('#'):
                t = t[1:]
            try:
                tickets[int(t)] = log.hexsha
            except ValueError:
                pass
        for b in lb:
            bugfixes[b.strip()] = log.hexsha
        if len(ls) == 0 and len(lf) == 0 and len(lt) == 0 and len(lb) == 0:
            changes[log.message] = log.hexsha


    #get ticket summaries from xmlrpc plugin installed on vistrails trac
    print "Will connect to VisTrails Trac with authentication..."
    if not username:
        print "Username:",
        sys.stdout.flush()
        username = raw_input()
        password = getpass.getpass()

    url = "https://%s:%s@www.vistrails.org/login/xmlrpc"%(username,
                                                               password)
    server = xmlrpclib.ServerProxy(url)
    print "downloading tickets.",
    for (tid,r) in tickets.items():
        print ".",
        sys.stdout.flush()
        try:
            ticket_info[tid] = server.ticket.get(tid)
        except Exception, e:
            del tickets[tid]
            print "commit %s: Could not get info for ticket %s"%(r,tid)
    print "done."

    #place tickets on bugfixes or enhancements
    for (tid,r) in tickets.iteritems():
        txt = "Ticket %s: %s"%(tid,ticket_info[tid][3]['summary'])
        if ticket_info[tid][3]['type'] == 'enhancement':
            features[txt] = r
        elif ticket_info[tid][3]['type'] in ['defect', 'defect+question']:
            bugfixes[txt] = r
        else:
            #put the rest as changes
            changes[txt] = r
    if commit_end == "HEAD" and len(logs) > 0:
        commit_end = logs[0].hexsha

    print
    print
    print "Release Name: v%s build %s from %s branch" % (release_name,
                                                         commit_end[0:12],
                                                         branch)
    print
    print "Enhancements:"
    times = []
    for t, r in features.iteritems():
        times.append((log_map_time[r], t))
    revisions = sorted(times)
    revisions.reverse()
    for (t,text) in revisions:
        r = features[text]
        print " - %s (%s)" %(rstrip(text),r[0:12])

    print
    print "Bug fixes:"
    times = []
    for t, r in bugfixes.iteritems():
        times.append((log_map_time[r], t))
    revisions = sorted(times)
    revisions.reverse()
    for (t,text) in revisions:
        r = bugfixes[text]
        print " - %s (%s)" %(rstrip(text),r[0:12])

    print
    print "Other changes:"
    times = []
    for t, r in changes.iteritems():
        times.append((log_map_time[r], t))
    revisions = sorted(times)
    revisions.reverse()
    for (t,text) in revisions:
        r = changes[text]
        print " - %s (%s)" %(text.split('\n')[0][0:100].rstrip(),r[0:12])

if __name__ == "__main__":
    repo = init_repo()
    build_release_notes(repo, branch)
    cleanup_repo()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
################################################################################
# ImageViewer widgets/toolbar implementation
################################################################################
from __future__ import division

import os
from PyQt4 import QtCore, QtGui
from vistrails.packages.spreadsheet.basic_widgets import SpreadsheetCell, SpreadsheetMode
from vistrails.packages.spreadsheet.spreadsheet_cell import QCellWidget, QCellToolBar
from vistrails.packages.spreadsheet.spreadsheet_controller import spreadsheetController
import imageviewer_rc

################################################################################

class ImageFileToSpreadsheet(SpreadsheetMode):
    def compute_output(self, output_module, configuration=None):
        fname = output_module.get_input('value').name
        window = spreadsheetController.findSpreadsheetWindow()
        local_file = window.file_pool.make_local_copy(fname)
        self.display_and_wait(output_module, configuration,
                              ImageViewerCellWidget, (local_file,))

class ImageViewerCell(SpreadsheetCell):
    """
    ImageViewerCell is a custom Module to display labels, images, etc.
    
    """    
    def compute(self):
        """ compute() -> None
        Dispatch the display event to the spreadsheet with images and labels
        
        """
        if self.has_input("File"):
            window = spreadsheetController.findSpreadsheetWindow()
            file_to_display = self.get_input("File")
            fileValue = window.file_pool.make_local_copy(file_to_display.name)
        else:
            fileValue = None
        self.displayAndWait(ImageViewerCellWidget, (fileValue, ))

class ImageViewerCellWidget(QCellWidget):
    """
    ImageViewerCellWidget is the actual QLabel that will draw
    labels/images on the spreadsheet
    
    """
    def __init__(self, parent=None):
        """ ImageViewerCellWidget(parent: QWidget) -> ImageViewerCellWidget
        Initialize the widget with its toolbar type and aligment
        
        """
        QCellWidget.__init__(self, parent)
        self.setLayout(QtGui.QVBoxLayout(self))
        self.setAnimationEnabled(True)
        
        self.label = QtGui.QLabel()
        self.layout().addWidget(self.label)
        
        self.label.setAutoFillBackground(True)
        self.label.palette().setColor(QtGui.QPalette.Window, QtCore.Qt.white)
        self.label.setMouseTracking(False)
        self.label.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter)
        self.label.setScaledContents(False)
        self.toolBarType = ImageViewerToolBar
        self.originalPix = None
        self.setMinimumSize(1, 1)

    def updateContents(self, inputPorts):
        """ updateContents(inputPorts: tuple) -> None
        Update the widget contents based on the input data
        
        """
        (fileValue, ) = inputPorts
        if fileValue:
            img = QtGui.QImage()
            if img.load(fileValue.name):
                self.originalPix = QtGui.QPixmap.fromImage(img)
                self.label.setPixmap(self.originalPix.scaled(self.label.size(),
                                                         QtCore.Qt.KeepAspectRatio,
                                                         QtCore.Qt.SmoothTransformation))
            else:
                self.label.setText("Invalid image file!")

        QCellWidget.updateContents(self, inputPorts)

    def saveToPNG(self, filename):
        """ saveToPNG(filename: str) -> bool
        Save the current widget contents to an image file
        
        """
        pixmap = self.label.pixmap()
        if pixmap and (not pixmap.isNull()):
            return pixmap.save(filename)
        return False

    def dumpToFile(self, filename):
        """ dumpToFile(filename: str) -> None
        Dumps the cell as an image file

        """
        pixmap = self.label.pixmap()
        if pixmap and (not pixmap.isNull()):
            if not os.path.splitext(filename)[1]:
                pixmap.save(filename, 'PNG')
            else:
                pixmap.save(filename)

    def saveToPDF(self, filename):
        """ saveToPDF(filename: str) -> bool
        Save the current widget contents to a pdf file
        
        """
        printer = QtGui.QPrinter()
        
        printer.setOutputFormat(QtGui.QPrinter.PdfFormat)
        printer.setOutputFileName(filename)
        painter = QtGui.QPainter()
        painter.begin(printer)
        rect = painter.viewport()
        pixmap = self.label.pixmap()
        size = pixmap.size()
        size.scale(rect.size(), QtCore.Qt.KeepAspectRatio)
        painter.setViewport(rect.x(), rect.y(), size.width(), size.height())
        painter.setWindow(pixmap.rect())
        painter.drawPixmap(0, 0, pixmap)
        painter.end()

    def resizeEvent(self, e):
        if self.originalPix!=None:
            self.label.setPixmap(self.originalPix.scaled(self.label.size(),
                                                         QtCore.Qt.KeepAspectRatio,
                                                         QtCore.Qt.SmoothTransformation))
                
class ImageViewerFitToCellAction(QtGui.QAction):
    """
    ImageViewerFitToCellAction is the action to stretch the image to
    fit inside a cell
    
    """
    def __init__(self, parent=None):
        """ ImageViewerFitToCellAction(parent: QWidget)
                                       -> ImageViewerFitToCellAction
        Setup the image, status tip, etc. of the action
        
        """
        QtGui.QAction.__init__(self,
                               QtGui.QIcon(":/images/fittocell.png"),
                               "&Fit To Cell",
                               parent)
        self.setStatusTip("Scale image content to fit cell frame")
        self.setCheckable(True)
        self.setChecked(True)

    def toggledSlot(self, checked):
        """ toggledSlot(checked: boolean) -> None
        Execute the action when the button is toggled
        
        """
        cellWidget = self.toolBar.getSnappedWidget()
        cellWidget.label.setScaledContents(checked)
        self.toolBar.slider.updateStatus((self.toolBar.sheet,
                                          self.toolBar.row,
                                          self.toolBar.col,
                                          cellWidget))
        
    def updateStatus(self, info):
        """ updateStatus(info: tuple) -> None
        Updates the status of the button based on the input info
        
        """
        (sheet, row, col, cellWidget) = info
        self.setChecked(cellWidget.label.hasScaledContents())


class ImageViewerZoomSlider(QtGui.QSlider):
    """
    ImageViewerZoomSlider is a slider that allows user to zoom in and
    out by dragging it
    
    """
    def __init__(self, parent=None):
        """ ImageViewerZoomSlider(parent: QWidget) -> ImageViewerZoomSlider
        Setup the ranges, status tip, etc. of the slider
        
        """
        QtGui.QSlider.__init__(self, QtCore.Qt.Horizontal, parent)
        self.setRange(100, 300)
        self.setValue(100)
        self.setTracking(True)
        self.setStatusTip("Zoom in the image")
        self.connect(self, QtCore.SIGNAL("valueChanged(int)"), self.updateZoom)
        self.connect(self, QtCore.SIGNAL("needUpdateStatus"), self.updateStatus)
        self.setSizePolicy(QtGui.QSizePolicy.Preferred,
                           QtGui.QSizePolicy.Expanding)
        
    def updateZoom(self, value):
        """ updateZoom(value: int) -> None
        Update the image when the slider value changed
        
        """
        if self.toolBar:
            cellWidget = self.toolBar.getSnappedWidget()
            if not cellWidget.label.hasScaledContents():
                newWidth = cellWidget.originalPix.width()*value//100
                pixmap = cellWidget.originalPix.scaledToWidth(newWidth)
                cellWidget.label.setPixmap(pixmap)

    def updateStatus(self, info):
        """ updateStatus(info: tuple) -> None
        Updates the status of the button based on the input info
        
        """
        (sheet, row, col, cellWidget) = info
        if cellWidget:
            if (not cellWidget.label.hasScaledContents() and
                not cellWidget._playing):
                self.setEnabled(True)
                originalWidth = cellWidget.originalPix.width()
                self.setValue(cellWidget.label.pixmap().width()*100//originalWidth)
            else:
                self.setEnabled(False)
                self.setValue(100)
                
class ImageViewerZoomLabel(QtGui.QLabel):
    """
    ImageViewerZoomLabel is the label sitting next to the ImageViewerZoomSlider
    
    """
    def __init__(self, parent=None):
        """ ImageViewerZoomSlider(parent: QWidget) -> None
        Setup the label with a status tip
        
        """
        QtGui.QLabel.__init__(self, "100%", parent)
        self.setStatusTip("Zoom in the image")
        
    def updateValue(self, value):
        """ updateValue(value: int)
        Updates the label with the new percentage value
        """
        self.setText(str(value)+"%")
                
class ImageViewerRotateAction(QtGui.QAction):
    """
    ImageViewerRotateAction is the action to rotate the image
    
    """
    def __init__(self, parent=None):
        """ ImageViewerRotateAction(parent: QWidget)
                                       -> ImageViewerRotateAction
        Setup the image, status tip, etc. of the action
        
        """
        QtGui.QAction.__init__(self,
                               QtGui.QIcon(":/images/rotate.png"),
                               "&Rotate CW...",
                               parent)
        self.setStatusTip("Rotate 90 degrees CW")
        self.rotationMatrix = QtGui.QMatrix(0,1,-1,0,0,0)
        
    def triggeredSlot(self, checked=False):
        """ toggledSlot(checked: boolean) -> None
        Execute the action when the button is clicked
        
        """
        cellWidget = self.toolBar.getSnappedWidget()
        if not cellWidget.label.pixmap() or cellWidget.label.pixmap().isNull():
            return
        cellWidget.originalPix = cellWidget.originalPix.transformed(
            self.rotationMatrix)
        cellWidget.label.setPixmap(cellWidget.label.pixmap().transformed(
            self.rotationMatrix))

class ImageViewerFlipAction(QtGui.QAction):
    """
    ImageViewerFlipAction is the action to flip the image
    
    """
    def __init__(self, parent=None):
        """ ImageViewerFlipAction(parent: QWidget) -> ImageViewerFlipAction
        Setup the image, status tip, etc. of the action
        
        """
        QtGui.QAction.__init__(self,
                               QtGui.QIcon(":/images/flip.png"),
                               "&Flip Horizontal...",
                               parent)
        self.setStatusTip("Flip the image horizontally")
        self.flipMatrix = QtGui.QMatrix(-1,0,0,1,0,0)
        
    def triggeredSlot(self, checked=False):
        """ toggledSlot(checked: boolean) -> None
        Execute the action when the button is clicked
        
        """
        cellWidget = self.toolBar.getSnappedWidget()
        label = cellWidget.label
        if not label.pixmap() or label.pixmap().isNull():
            return
        cellWidget.originalPix = cellWidget.originalPix.transformed(
            self.flipMatrix)
        label.setPixmap(label.pixmap().transformed(self.flipMatrix))

class ImageViewerToolBar(QCellToolBar):
    """
    ImageViewerToolBar derives from CellToolBar to give the ImageViewerCellWidget
    a customizable toolbar
    
    """
    def createToolBar(self):
        """ createToolBar() -> None
        This will get call initiallly to add customizable widgets
        
        """
        self.appendAction(ImageViewerFitToCellAction(self))
        self.appendAction(ImageViewerRotateAction(self))
        self.appendAction(ImageViewerFlipAction(self))
        self.slider = ImageViewerZoomSlider(self)
        label = ImageViewerZoomLabel(self)
        self.connect(self.slider,
                     QtCore.SIGNAL("valueChanged(int)"),
                     label.updateValue)
        self.appendWidget(self.slider)
        self.appendWidget(label)
        self.addAnimationButtons()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
"""
This module defines common functions and exception class definitions
used all over VisTrails.
"""
from __future__ import division, with_statement

from vistrails.core.utils.enum import enum
from vistrails.core.utils.timemethod import time_method, time_call
from vistrails.core.utils.tracemethod import trace_method, bump_trace, report_stack, \
     trace_method_options, trace_method_args
from vistrails.core.utils.color import ColorByName
import copy
from distutils.version import LooseVersion
import errno
import functools
import itertools
import os
import sys
import warnings
import weakref


import unittest
import tempfile

################################################################################

def invert(d):
    """invert(dict) -> dict. 

    Returns an inverted dictionary by switching key-value pairs. If
    you use this repeatedly, consider switching the underlying data
    structure to a core.data_structures.bijectivedict.Bidict instead.

    """
    return dict([[v,k] for k,v in d.items()])

################################################################################

def unimplemented():
    """Raises UnimplementedException."""
    raise UnimplementedException()

def abstract():
    """Raises AbstractException.""" 
    raise AbstractException()

class VistrailsWarning(Warning):
    pass

class VistrailsDeprecation(VistrailsWarning):
    pass

def deprecated(*args):
    new_name = None
    def _deprecated(func):
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            if new_name is not None:
                warnings.warn("Call to deprecated function %s "
                              "replaced by %s" % (
                                  func.__name__, new_name),
                              category=VistrailsDeprecation,
                              stacklevel=2)
            else:
                warnings.warn("Call to deprecated function %s" % func.__name__,
                              category=VistrailsDeprecation,
                              stacklevel=2)
            return func(*args, **kwargs)
        return new_func
    if len(args) == 1 and callable(args[0]):
        return _deprecated(args[0])
    else:
        new_name = args[0]
        return _deprecated

################################################################################

class NoMakeConnection(Exception):
    """NoMakeConnection is raised when a VisConnection doesn't know how to
    create a live version of itself. This is an internal error that
    should never be seen by a user. Please report a bug if you see
    this.

    """
    def __init__(self, conn):
        self.conn = conn
    def __str__(self):
        return "Connection %s has no makeConnection method" % self.conn

class NoSummon(Exception):
    """NoSummon is raised when a VisObject doesn't know how to create a
    live version of itself. This is an internal error that should
    never be seen by a user. Please report a bug if you see this.

    """
    def __init__(self, obj):
        self.obj = obj
    def __str__(self):
        return "Module %s has no summon method" % self.obj

class UnimplementedException(Exception):
    """UnimplementedException is raised when some interface hasn't been
    implemented yet. This is an internal error that should never be
    seen by a user. Please report a bug if you see this.

    """
    def __str__(self):
        return "Object is Unimplemented"

class AbstractException(Exception):
    """AbstractException is raised when an abstract method is called.
    This is an internal error that should never be seen by a
    user. Please report a bug if you see this.

    """
    def __str__(self):
        return "Abstract Method was called"

class VistrailsInternalError(Exception):
    """VistrailsInternalError is raised when an unexpected internal
    inconsistency happens. This is (clearly) an internal error that
    should never be seen by a user. Please report a bug if you see
    this.

    """
    def __str__(self):
        return "Vistrails Internal Error: " + str(self.message)

class VersionTooLow(Exception):
    """VersionTooLow is raised when you're running an outdated version of
    some necessary software or package.

    """
    def __init__(self, sw, required_version):
        self.sw = sw
        self.required_version = required_version
    def __str__(self):
        return ("Your version of '" +
                self.sw +
                "' is too low. Please upgrade to " +
                self.required_version +
                " or later")

class InvalidModuleClass(Exception):
    """InvalidModuleClass is raised when there's something wrong with a
    class that's being registered as a module within VisTrails.

    """

    def __init__(self, klass):
        self.klass = klass

    def __str__(self):
        return ("class '%s' cannot be registered in VisTrails. Please" +
                " consult the documentation.") % self.klass.__name__

class ModuleAlreadyExists(Exception):
    """ModuleAlreadyExists is raised when trying to add a class that
    is already in the module registry."""

    def __init__(self, identifier, moduleName):
        self._identifier = identifier
        self._name = moduleName

    def __str__(self):
        return ("'%s, %s' cannot be registered in VisTrails because of another "
                "module with the same identifier and name already exists." %
                (self._identifier,
                 self._name))

class PortAlreadyExists(Exception):
    """PortAlreadyExists is raised when trying to add a PortSpec that
    has the same name and type as an existing PortSpec."""

    def __init__(self, identifier, module_name, port_type, port_name):
        self._identifier = identifier
        self._module_name = module_name
        self._port_type = port_type
        self._port_name = port_name
        
    def __str__(self):
        return "Module '%s:%s' already contains an %s port named '%s'" % \
            (self._identifier, self._module_name, self._port_type,
             self._port_name)

class InvalidPipeline(Exception):
    """InvalidPipeline is raised when a pipeline cannot be instantiated due 
    to missing information in the registry, like unloaded packages or missing
    modules.

    parameters:

    exception_set: list of all exceptions related to why this is an
    invalid pipeline

    pipeline: pipeline that is invalid (potentially incomplete, in the
    case where modules could not be found, etc). This is stored here
    so that pipeline upgrades can be performed appropriately. Since
    Controller.do_version_switch (sensibly) bails before setting the
    invalid pipeline to current_pipeline and the new value to
    current_version, these need to percolate through the exceptions so
    the values can be fixed.

    version: version id of the pipeline in the vistrail
    """
    def __init__(self, exception_set, pipeline=None, version=None):
        self._exception_set = exception_set
        #the problem here is that sometimes the pipeline can't be copied when
        # it is invalid. So if it throws an Exception, we will just ignore
        try:
            self._pipeline = copy.copy(pipeline)
        except Exception:
            self._pipeline = None
        self._version = version

    def __str__(self):
        return "Pipeline has errors. Please see the detailed message for more information.\n  " + \
            '\n  '.join(line for e in self._exception_set 
                        for line in str(e).splitlines())

    def get_exception_set(self):
        return self._exception_set

################################################################################

# Only works for functions with NO kwargs!
def memo_method(method):
    """memo_method is a method decorator that memoizes results of the
    decorated method, trading off memory for time by caching previous
    results of the calls."""
    attrname = "_%s_memo_result" % id(method)
    memo = {}
    def decorated(self, *args):
        try:
            return memo[args]
        except KeyError:
            result = method(self, *args)
            memo[args] = result
            return result
    warn = "(This is a memoized method: Don't mutate the return value you're given.)"
    if method.__doc__:
        decorated.__doc__ = method.__doc__ + "\n\n" + warn
    else:
        decorated.__doc__ = warn
    return decorated

##############################################################################
# Profiling, utilities

_profiled_list = []
def profile(func):
    """profile is a method decorator that profiles the calls of a
    given method using cProfile. You need to get the decorated method
    programmatically later to get to the profiler stats. It will be
    available as the attribute 'profiler_object' on the decorated
    result.

    From there, you can simply call save_all_profiles(), and that will
    take the list of all profiled methods and save them to different
    files.

    If you like manual labor, you probably want to do something like this:

    >>> po = ...... .profiler_object
    >>> po.dump_stats('/tmp/some_temporary_file')
    >>> import pstats
    >>> ps = pstats.Stats('/tmp/some_temporary_file')
    >>> ps.sort_stats('time') # or cumtime, or calls, or others - see doc
    >>> ps.print_stats()

    """

    # Notice that on ubuntu you will need
    # sudo apt-get install python-profiler

    try:
        import cProfile as prof
    except ImportError:
        import profile as prof

    pobject = prof.Profile()
    def method(*args, **kwargs):
        return pobject.runcall(func, *args, **kwargs)

    method.profiler_object = pobject
    _profiled_list.append((func.__name__, method))
    return method

def get_profiled_methods():
    return _profiled_list

def save_profile_to_disk(callable_, filename):
    callable_.profiler_object.dump_stats(filename)

def save_all_profiles():
    td = tempfile.gettempdir()
    for (name, method) in get_profiled_methods():
        fout = td + name + '.pyp'
        #print fout
        method.profiler_object.dump_stats(fout)

##############################################################################

def debug(func):
    """debug is a method decorator that invokes the python integrated
    debugger in a given method. Use it to step through tricky
    code. Note that pdb is not integrated with emacs or anything like
    that, so you'll need a shell to see what's going on.

    """
    import pdb

    def method(*args, **kwargs):
        return pdb.runcall(func, *args, **kwargs)
    return method

################################################################################

# Write our own all() and any() if python version < 2.5
if sys.version_info < (2, 5):
    def any(iterable):
        """any(iterable) -> Boolean - Returns true if any element
        is true. This is meant to be the equivalent of python 2.5's any
        when running on python < 2.5"""
        for b in iterable:
            if b:
                return True
        return False

    def all(iterable):
        """all(iterable) -> Boolean - Returns true if no elements are
        False.  This is meant to be the equivalent of python 2.5's
        all() when running on python < 2.5"""
        for b in iterable:
            if not b:
                return False
        return True
else:
    import __builtin__
    any = __builtin__.any
    all = __builtin__.all

def iter_index(iterable, item):
    """iter_index(iterable, item) -> int - Iterates through iterator
    until item is found, and returns the index inside the iterator.

    iter_index is analogous to list.index for iterators."""
    try:
        itor = itertools.izip(iterable, itertools.count(0))
        return itertools.dropwhile(lambda (v,c): v != item, itor).next()[1]
    except StopIteration:
        return -1
                                              

def eprint(*args):
    """eprint(*args) -> False - Prints the arguments, then returns
    false. Useful inside a lambda expression, for example."""
    for v in args:
        print v,
    print

def uniq(l):
    """uniq(l) -> List. Returns a new list consisting of elements that
    test pairwise different for equality. Requires all elements to be
    sortable, and runs in O(n log n) time."""
    if len(l) == 0:
        return []
    a = copy.copy(l)
    a.sort()
    l1 = a[:-1] 
    l2 = a[1:]
    return [a[0]] + [next for (i, next) in itertools.izip(l1, l2) if i != next]

class InstanceObject(object):
    """InstanceObject is a convenience class created to facilitate
    creating of one-off objects with many fields. It simply translates
    the passed kwargs on the constructor to a set of fields with
    the right values."""
    def __init__(self, **kw):
        self.__dict__.update(kw)

    def __str__(self):
        pre = "(%s " % self.__class__.__name__
        items = [('%s: %s' % (k, str(v)))
                 for (k, v)
                 in sorted(self.__dict__.items())]
        items_str = ('\n' + (' ' * len(pre))).join(items)
        post = ')@%X' % id(self)
        return pre + items_str + post
    
    def write_source(self, prefix=""):
        result = ""
        for (k, v) in sorted(self.__dict__.items()):
            if isinstance(v, InstanceObject):
                newprefix = prefix + "." + k
                result += v.write_source(newprefix)
            else:
                result += prefix
                result += "." + str(k) + " = " 
                if isinstance(v, basestring):
                    result +=  "'" + str(v) + "'\n"
                else:
                    result += str(v) + "\n"
        return result

def append_to_dict_of_lists(dict, key, value):
    """Appends /value/ to /dict/[/key/], or creates entry such that
    /dict/[/key/] == [/value/]."""
    try:
        dict[key].append(value)
    except KeyError:
        dict[key] = [value]

def version_string_to_list(version):
    """version_string_to_list converts a version string to a list of
    numbers and strings:

    version_string('0.1') -> [0, 1]
    version_string('0.9.9alpha1') -> [0, 9, 9, alpha', 1]

    """
    return LooseVersion(version).version

def versions_increasing(v1, v2):
    return LooseVersion(v1) < LooseVersion(v2)

##############################################################################
# DummyView & DummyScene

class DummyScene(object):
    def __init(self):
        self.current_version = -1
        self.current_pipeline = None
    def get_selected_module_ids(self):
        return [] 
    def flushMoveActions(self, *args, **kwargs): pass
    
class DummyView(object):
    def __init__(self):
        self._scene = DummyScene()
    def set_module_active(self, *args, **kwargs): pass
    def set_module_computing(self, *args, **kwargs): pass
    def set_module_success(self, *args, **kwargs): pass
    def set_module_suspended(self, *args, **kwargs): pass
    def set_module_error(self, *args, **kwargs): pass
    def set_module_not_executed(self, *args, **kwargs): pass
    def set_module_progress(self, *args, **kwargs): pass
    def set_module_persistent(self, *args, **kwargs): pass
    def set_execution_progress(self, *args, **kwargs): pass
    def flushMoveActions(self, *args, **kwargs): pass
    def scene(self): 
        return self._scene
##############################################################################    
# FIXME: Add tests
def no_interrupt(callable_, *args, **kwargs):
    """no_interrupt(callable_, *args, **kwargs) -> return arguments
    from callable.

    Calls callable_ with *args and **kwargs and keeps retrying as long as call
is interrupted by the OS. This makes calling read more convenient when
using output from the subprocess module."""
    while True:
        try:
            return callable_(*args, **kwargs)
        except IOError, e:
            if e.errno == errno.EINTR:
                continue
            else:
                raise

################################################################################
# class for creating weak references to bound methods
# based on recipe http://code.activestate.com/recipes/81253/
# converted to work also in python 2.6.x without using deprecated methods
# not tested in python 2.5.x but it should work
class Ref(object):
    """ Wraps any callable, most importantly a bound method, in a way that
    allows a bound method's object to be GC'ed, while providing the same
    interface as a normal weak reference. """
    def __init__(self, fn):
        try:
            #try getting object, function, and class
            o, f, c = fn.im_self, fn.im_func, fn.im_class
        except AttributeError: #it's not a bound method
            self._obj = None
            self._func = fn
            self._clas = None
        else: #it's a bound method
            if o is None: self._obj = None #... actually UN-bound
            else: self._obj = weakref.ref(o)
            self._func = f
            self._clas = None
            
    def __call__(self):
        if self._obj is None: return self._func
        elif self._obj() is None: return None
        try:
            import types
            instance_method = types.MethodType
        except ImportError:
            #new is deprecated in python 2.6
            import new
            instance_method = new.instancemethod
        return instance_method(self._func, self._obj(), self._clas)

###############################################################################

def xor(first, *others):
    """XORs bytestrings.

    Example: xor('abcd', '\x20\x01\x57\x56') = 'Ac42'
    """
    l = len(first)
    first = [ord(c) for c in first]
    for oth in others:
        if len(oth) != l:
            raise ValueError("All bytestrings should have the same length: "
                             "%d != %d" % (l, len(oth)))
        first = [c ^ ord(o) for (c, o) in itertools.izip(first, oth)]
    return ''.join(chr(c) for c in first)

def long2bytes(nb, length=None):
    """Turns a single integer into a little-endian bytestring.

    Uses as many bytes as necessary or optionally pads to length bytes.
    Might return a result longer than length.

    Example: long2bytes(54321, 4) = b'\x31\xD4\x00\x00'
    """
    if nb < 0:
        raise ValueError
    elif nb == 0:
        result = b'\x00'
    else:
        result = b''
        while nb > 0:
            result += chr(nb & 0xFF)
            nb = nb >> 8
    if length is not None and len(result) < length:
        result += '\x00' * (length - len(result))
    return result

################################################################################

class Chdir(object):
    def __init__(self, dirname):
        self._old_dir = os.getcwd()
        self._new_dir = dirname
        
    def __enter__(self):
        os.chdir(self._new_dir)
        
    def __exit__(self, *args):
        os.chdir(self._old_dir)
################################################################################



class _TestRegularFibo(object):
    def __init__(self):
        self.calls = 0

    def f(self, x):
        self.calls += 1
        if x == 0: return 0
        if x == 1: return 1
        return self.f(x-1) + self.f(x-2)

class _TestMemoFibo(_TestRegularFibo):
    f = memo_method(_TestRegularFibo.f)

class TestCommon(unittest.TestCase):
    def test_append_to_dict_of_lists(self):
        f = {}
        self.assertEquals(f.has_key(1), False)
        append_to_dict_of_lists(f, 1, 1)
        self.assertEquals(f.has_key(1), True)
        self.assertEquals(f[1], [1])
        append_to_dict_of_lists(f, 1, 1)
        self.assertEquals(f.has_key(1), True)
        self.assertEquals(f[1], [1, 1])
        append_to_dict_of_lists(f, 1, 2)
        self.assertEquals(f.has_key(1), True)
        self.assertEquals(f[1], [1, 1, 2])
        append_to_dict_of_lists(f, 2, "Foo")
        self.assertEquals(f.has_key(2), True)
        self.assertEquals(f[2], ["Foo"])

    def test_memo(self):
        regular = _TestRegularFibo()
        r1 = regular.f(20)
        memoized = _TestMemoFibo()
        r2 = memoized.f(20)
        self.assertEqual(r1, 6765)
        self.assertEqual(r2, 6765)
        self.assertLess(memoized.calls, regular.calls)
        self.assertEqual(regular.calls, 21891)
        self.assertEqual(memoized.calls, 21)

    def test_memo_2(self):
        count = [0]
        class C1(object):
            pass
        class C2(object):
            pass
        class TestClassMemo(object):
            def __init__(self, cell):
                self.cell = cell
            @memo_method
            def f(self, cl, x):
                self.cell[0] += 1
                return x
        t = TestClassMemo(count)
        self.assertEquals(count[0], 0)
        t.f(C1, 0)
        self.assertEquals(count[0], 1)
        t.f(C1, 0)
        self.assertEquals(count[0], 1)
        t.f(C1, 1)
        self.assertEquals(count[0], 2)
        t.f(C2, 0)
        self.assertEquals(count[0], 3)

    def test_version_string_to_list(self):
        self.assertEquals(version_string_to_list("0.1"), [0, 1])
        self.assertEquals(version_string_to_list("1.0.2"), [1, 0, 2])
        self.assertEquals(version_string_to_list("1.0.2beta"),
                          [1, 0, 2, 'beta'])
    
    def test_ref(self):
        class C(object):
            def f(self):
                return 'hello'
        c = C()
        cf = weakref.ref(c.f)
        #bound methods behave not as expected. You want cf() not to be None
        self.assertEquals(cf(), None)
        #so we use the new class
        cf = Ref(c.f)
        #it behaves as expected
        self.assertEquals(cf()(),'hello')
        del c
        #and after deletion the reference is dead
        self.assertEquals(cf(), None)
        
    def test_chdir(self):
        def raise_exception():
            with Chdir(tempfile.gettempdir()):
                raise Exception
            
        currentpath = os.getcwd()
        with Chdir(tempfile.gettempdir()):
            pass
        self.assertEquals(os.getcwd(), currentpath)
        
        self.assertRaises(Exception, raise_exception)
        self.assertEquals(os.getcwd(), currentpath)

    def test_deprecated(self):
        import re
        def canon_path(path):
            path = os.path.realpath(path)
            p, f = os.path.dirname(path), os.path.basename(path)
            f = re.split(r'[$.]', f)[0]
            return os.path.join(p, f)
        def check_warning(msg, f):
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter('always')
                f(1, 2)
            self.assertEqual(len(w), 1)
            w, = w
            self.assertEqual(w.message.message, msg)
            self.assertEqual(w.category, VistrailsDeprecation)
            self.assertTrue(canon_path(w.filename),
                            canon_path(__file__))

        @deprecated('repl1')
        def func1(a, b):
            self.assertEqual((a, b), (1, 2))
        @deprecated
        def func2(a, b):
            self.assertEqual((a, b), (1, 2))
        check_warning('Call to deprecated function func1 replaced by repl1',
                      func1)
        check_warning('Call to deprecated function func2', func2)

        foo = None
        class Foo(object):
            @deprecated('repl1')
            def meth1(s, a, b):
                self.assertEqual((s, a, b), (foo, 1, 2))
            @deprecated
            def meth2(s, a, b):
                self.assertEqual((s, a, b), (foo, 1, 2))
            @staticmethod
            @deprecated('repl3')
            def meth3(a, b):
                self.assertEqual((a, b), (1, 2))
            @staticmethod
            @deprecated
            def meth4(a, b):
                self.assertEqual((a, b), (1, 2))
        foo = Foo()
        check_warning('Call to deprecated function meth1 replaced by repl1',
                      foo.meth1)
        check_warning('Call to deprecated function meth2',
                      foo.meth2)
        check_warning('Call to deprecated function meth3 replaced by repl3',
                      foo.meth3)
        check_warning('Call to deprecated function meth4',
                      foo.meth4)


if __name__ == '__main__':
    unittest.main()



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2013-2014, NYU-Poly.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

from vistrails.core.modules.basic_modules import Constant, String
from vistrails.core.modules.config import IPort, OPort
from vistrails.core.modules.vistrails_module import ModuleError


@apply
class StoreHolder(object):
    def __init__(self):
        self.store = None

    def get_store(self):
        return self.store

    def set_store(self, store):
        self.store = store

set_default_store = StoreHolder.set_store
get_default_store = StoreHolder.get_store


# The type of the file, i.e. how VisTrails stored it
KEY_TYPE = 'vistrails_objecttype'
TYPE_CACHED =   'cached'    # A cached path, with very little metadata
TYPE_INPUT =    'input'     # An input file, i.e. an external file that
                            # VisTrails added to the store as-is
TYPE_OUTPUT =   'output'    # An intermediate or output file, i.e. a file that
                            # was generated by VisTrails

# Timestamp for file insertion
KEY_TIME = 'vistrails_timestamp'

# Signature of the module that added the file (cached and output, not input)
KEY_SIGNATURE = 'vistrails_signature'

# Identifier for the workflow
KEY_WORKFLOW = 'vistrails_workflow'

# Module ID in the workflow
KEY_MODULE_ID = 'vistrails_module_id'


class PersistentHash(Constant):
    """Reference to a specific file.

    Unequivocally references a specific file (by its full hash).
    """
    def __init__(self, h=None):
        Constant.__init__(self)
        if h is not None:
            self._set_hash(h)
        else:
            self._hash = None

    def _set_hash(self, h):
        if not (isinstance(h, basestring)):
            raise TypeError("File hash should be a string")
        elif len(h) != 40:
            raise ValueError("File hash should be 40 characters long")
        if not isinstance(h, bytes):
            h = bytes(h)
        self._hash = h

    @staticmethod
    def translate_to_python(h):
        try:
            return PersistentHash(h)
        except (TypeError, ValueError):
            return None

    @staticmethod
    def translate_to_string(ref):
        if ref._hash is not None:
            return ref._hash
        else:
            raise ValueError("Reference is invalid")

    @staticmethod
    def validate(ref):
        return isinstance(ref, PersistentHash)

    def __str__(self):
        return self._hash

    def __repr__(self):
        if self._hash is not None:
            return "<PersistentHash %s>" % self._hash
        else:
            return "<PersistentHash (invalid)>"

    def compute(self):
        if self.has_input('value') == self.has_input('hash'):
            raise ModuleError(self, "Set either 'value' or 'hash'")
        if self.has_input('value'):
            self._hash = self.get_input('value')._hash
        else:
            try:
                self._set_hash(self.get_input('hash'))
            except ValueError, e:
                raise ModuleError(self, e.message)

PersistentHash._input_ports = [
        IPort('value', PersistentHash),
        IPort('hash', String)]
PersistentHash._output_ports = [
        OPort('value', PersistentHash)]
